"""Sodarcache API for adding and updating cache items"""

import logging

from typing import Optional

from django.contrib.auth import get_user_model
from django.db.models import QuerySet

# Projectroles dependency
from projectroles.models import Project, SODARUser
from projectroles.plugins import PluginAPI
from projectroles.utils import get_app_names

from sodarcache.models import JSONCacheItem


logger = logging.getLogger(__name__)
plugin_api = PluginAPI()
User = get_user_model()


# Local variables
APP_NAMES = get_app_names()
LABEL_MAX_WIDTH = 32
CACHE_TYPES = ['json']


class SODARCacheAPI:
    """SODARCache backend API to be used by Django apps."""

    # TODO: Make model selection dynamic once we introduce types other than JSON

    # Internal functions -------------------------------------------------------

    @classmethod
    def _check_app_name(cls, app_name: str):
        """Check if app_name is valid, raise ValueError if not"""
        if app_name not in APP_NAMES:
            raise ValueError(
                'Unknown app name "{}" (installed apps: {})'.format(
                    app_name, ', '.join(x for x in APP_NAMES)
                )
            )

    @classmethod
    def _check_data_type(cls, data_type: str):
        """Check if data_type is valid, raise ValueError if not"""
        if data_type not in CACHE_TYPES:
            raise ValueError(
                'Unknown data type "{}" for a cache item '
                '(allowed types: {})'.format(
                    data_type, ', '.join(x for x in CACHE_TYPES)
                )
            )

    # API functions ------------------------------------------------------------

    @classmethod
    def get_model(cls) -> type[JSONCacheItem]:
        """
        Return JSONCacheItem model for direct model access.

        :returns: JSONCacheItem class
        """
        return JSONCacheItem

    @classmethod
    def get_project_cache(
        cls, project: Project, data_type: str = 'json'
    ) -> QuerySet:
        """
        Return all cached data for a project.

        :param project: Project object
        :param data_type: String stating the data type of the cache items
        :return: QuerySet
        :raise: ValueError if data_type is invalid
        """
        cls._check_data_type(data_type)
        return JSONCacheItem.objects.filter(project=project)

    @classmethod
    def update_cache(
        cls,
        name: Optional[str] = None,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ):
        """
        Update items by certain name within a project by calling implemented
        functions in project app plugins.

        :param name: Item name to limit update to (string, optional)
        :param project: Project object to limit update to (optional)
        :param user: User object to denote user triggering the update (optional)
        """
        plugins = plugin_api.get_active_plugins(plugin_type='project_app')
        for plugin in plugins:
            plugin.update_cache(name, project, user)

    @classmethod
    def delete_cache(
        cls, app_name: Optional[str] = None, project: Optional[Project] = None
    ) -> int:
        """
        Delete cache items. Optionallly limit to project and/or user.

        :param app_name: Name of the app which sets the item (string)
        :param project: Project object (optional)
        :return: Integer (deleted item count)
        :raise: ValueError if app_name is given but invalid
        """
        if app_name:
            cls._check_app_name(app_name)
        if not app_name and not project:
            items = JSONCacheItem.objects.all()
        else:
            query_params = {}
            if app_name:
                query_params['app_name'] = app_name
            if project:
                query_params['project'] = project
            items = JSONCacheItem.objects.filter(**query_params)

        if items:
            item_count = items.count()
            items.delete()
            logger.info(
                'Deleted {} item{} from cache(app={}, project={})'.format(
                    item_count,
                    's' if item_count != 1 else '',
                    app_name,
                    project.sodar_uuid if project else None,
                )
            )
            return item_count

        logger.info(
            'No items found for deletion (app={}, project={})'.format(
                app_name, project.sodar_uuid if project else None
            )
        )
        return 0

    @classmethod
    def get_cache_item(
        cls, app_name: str, name: str, project: Optional[Project] = None
    ) -> JSONCacheItem:
        """
        Return cached data by app_name, name (identifier) and optional project.
        Returns None if not found.

        :param name: Item name (string)
        :param app_name: Name of the app which sets the item (string)
        :param project: Project object (optional)
        :return: JSONCacheItem object
        :raise: ValueError if app_name is invalid
        """
        cls._check_app_name(app_name)
        query_string = {'app_name': app_name, 'name': name}
        if project:
            query_string['project'] = project
        return JSONCacheItem.objects.filter(**query_string).first()

    @classmethod
    def set_cache_item(
        cls,
        app_name: str,
        name: str,
        data: dict,
        data_type: str = 'json',
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ) -> JSONCacheItem:
        """
        Create or update and save a cache item.

        :param app_name: Name of the app which sets the item (string)
        :param name: Item name (string)
        :param data: Item data (dict)
        :param data_type: String stating the data type of the cache items
        :param project: Project object (optional)
        :param user: User object to denote user triggering the update (optional)
        :return: JSONCacheItem object
        :raise: ValueError if app_name is invalid
        :raise: ValueError if data_type is invalid
        """
        cls._check_app_name(app_name)
        cls._check_data_type(data_type)
        item = cls.get_cache_item(app_name, name, project)
        log_msg = f'Updated item "{app_name}:{name}"'
        if not item:
            if data_type == 'json':
                item = JSONCacheItem()
                item.name = name
                item.app_name = app_name
        item.data = data
        if project:
            item.project = project
            log_msg += f' in project {project.get_log_title()}'
        if user:
            item.user = user
            log_msg += f' by user "{user.username}"'
        item.save()
        logger.info(log_msg)
        return item

    @classmethod
    def delete_cache_item(
        cls, app_name: str, name: str, project: Optional[Project] = None
    ):
        """
        Method for deleting a cache item.

        :param app_name: Name of the app which sets the item (string)
        :param name: Item name (string)
        :param project: Project object (optional)
        """
        item = cls.get_cache_item(app_name, name, project)
        if item:
            item.delete()
            logger.info(f'Deleted item "{app_name}:{name}" from cache')

    @classmethod
    def get_update_time(
        cls, app_name: str, name: str, project: Optional[Project] = None
    ) -> float:
        """
        Return the time of the last update of a cache object as seconds since
        epoch.

        :param name: Item name (string)
        :param app_name: Name of the app which sets the item (string)
        :param project: Project object (optional)
        :return: Float
        """
        item = cls.get_cache_item(app_name, name, project)
        return item.date_modified.timestamp() if item else None



"""Plugin tests for the timeline app"""

from test_plus.test import TestCase

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import (
    ProjectAppPluginPoint,
    BackendPluginPoint,
    SiteAppPluginPoint,
    PluginSearchResult,
    PluginAPI,
)
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
)

from timeline.api import TimelineAPI
from timeline.models import TimelineEvent, TL_STATUS_OK
from timeline.plugins import STATS_DESC_USER_COUNT

# from timeline.tests.test_models import TimelineEventMixin
from timeline.urls import urls_ui_project, urls_ui_site, urls_ui_admin


plugin_api = PluginAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
PROJECT_PLUGIN_NAME = 'timeline'
PROJECT_PLUGIN_TITLE = 'Timeline'
BACKEND_PLUGIN_NAME = 'timeline_backend'
BACKEND_PLUGIN_TITLE = 'Timeline Backend'
SITE_PLUGIN_NAME = 'timeline_site'
SITE_PLUGIN_TITLE = 'Site-Wide Events'
ADMIN_PLUGIN_NAME = 'timeline_site_admin'
ADMIN_PLUGIN_TITLE = 'All Timeline Events'
EVENT_NAME = 'test_event'
SEARCH_TERMS = ['test']
SEARCH_RET_CAT = 'all'
SEARCH_RET_TITLE = 'Timeline Events'
SEARCH_RET_TYPES = ['timeline']


class TimelinePluginTestBase(
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    TestCase,
):
    """Base class for timeline plugin tests"""

    def setUp(self):
        # Init users
        self.user = self.make_user('superuser')
        self.user.is_staff = True
        self.user.is_superuser = True
        self.user.save()
        self.user_owner = self.make_user('user_owner')
        # Init roles
        self.init_roles()
        # Init category, project and roles
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as = self.make_assignment(
            self.category, self.user_owner, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )


class TestProjectAppPlugin(TimelinePluginTestBase):
    """Tests for timeline ProjectAppPlugin"""

    def _add_event(self) -> TimelineEvent:
        """Helper for adding event with OK status"""
        event = self.timeline.add_event(**self.event_kw)
        event.set_status(TL_STATUS_OK)
        return event

    def setUp(self):
        super().setUp()
        self.plugin = ProjectAppPluginPoint.get_plugin(PROJECT_PLUGIN_NAME)
        self.event_kw = {
            'project': self.project,
            'app_name': 'projectroles',
            'user': self.user,
            'event_name': EVENT_NAME,
            'description': 'description',
        }
        self.timeline = plugin_api.get_backend_api('timeline_backend')

    def test_plugin_retrieval(self):
        """Test retrieving ProjectAppPlugin"""
        self.assertIsNotNone(self.plugin)
        self.assertEqual(self.plugin.get_model().name, PROJECT_PLUGIN_NAME)
        self.assertEqual(self.plugin.name, PROJECT_PLUGIN_NAME)
        self.assertEqual(self.plugin.get_model().title, PROJECT_PLUGIN_TITLE)
        self.assertEqual(self.plugin.urls, urls_ui_project)

    def test_get_statistics(self):
        """Test get_statistics() with no events"""
        expected = {
            'event_count': {
                'label': 'Events',
                'value': 0,
            },
            'user_count': {
                'label': 'Users',
                'description': STATS_DESC_USER_COUNT,
                'value': 0,
            },
        }
        self.assertEqual(self.plugin.get_statistics(), expected)

    def test_get_statistics_events(self):
        """Test get_statistics() with existing events"""
        self._add_event()
        self.event_kw['user'] = self.user_owner
        self._add_event()
        ret = self.plugin.get_statistics()
        self.assertEqual(ret['event_count']['value'], 2)
        self.assertEqual(ret['user_count']['value'], 2)

    def test_get_statistics_same_user(self):
        """Test get_statistics() with existing events by same user"""
        self._add_event()
        self._add_event()
        ret = self.plugin.get_statistics()
        self.assertEqual(ret['event_count']['value'], 2)
        self.assertEqual(ret['user_count']['value'], 1)

    def test_get_statistics_no_user(self):
        """Test get_statistics() with existing events including no user"""
        self._add_event()
        self.event_kw['user'] = None
        self._add_event()
        ret = self.plugin.get_statistics()
        self.assertEqual(ret['event_count']['value'], 2)
        self.assertEqual(ret['user_count']['value'], 1)

    def test_search(self):
        """Test search() with no events"""
        ret = self.plugin.search(SEARCH_TERMS, self.user)
        self.assertEqual(len(ret), 1)
        self.assertIsInstance(ret[0], PluginSearchResult)
        self.assertEqual(ret[0].category, SEARCH_RET_CAT)
        self.assertEqual(ret[0].title, SEARCH_RET_TITLE)
        self.assertEqual(ret[0].search_types, SEARCH_RET_TYPES)
        self.assertEqual(ret[0].items, [])

    def test_search_events(self):
        """Test search() with events"""
        event = self._add_event()
        self.event_kw['user'] = self.user_owner
        event2 = self._add_event()
        ret = self.plugin.search(SEARCH_TERMS, self.user)
        self.assertEqual(len(ret), 1)
        self.assertIsInstance(ret[0].items, list)
        self.assertEqual(len(ret[0].items), 2)
        self.assertEqual(ret[0].items[0], event2)
        self.assertEqual(ret[0].items[1], event)

    def test_search_event_name(self):
        """Test search() with event name"""
        event = self._add_event()
        ret = self.plugin.search([EVENT_NAME], self.user)
        self.assertEqual(len(ret), 1)
        self.assertEqual(len(ret[0].items), 1)
        self.assertEqual(ret[0].items[0], event)

    def test_search_event_name_display(self):
        """Test search() with event name in display formatting"""
        event = self._add_event()
        ret = self.plugin.search(['Test Event'], self.user)
        self.assertEqual(len(ret), 1)
        self.assertEqual(len(ret[0].items), 1)
        self.assertEqual(ret[0].items[0], event)

    def test_search_invalid_terms(self):
        """Test search() with invalid terms"""
        self._add_event()
        ret = self.plugin.search(['yuyaeQu7ma6aeFi2'], self.user)
        self.assertEqual(len(ret[0].items), 0)

    def test_search_mixed_terms(self):
        """Test search() with valid and invalid terms"""
        self._add_event()
        ret = self.plugin.search(SEARCH_TERMS + ['yuyaeQu7ma6aeFi2'], self.user)
        self.assertEqual(len(ret[0].items), 1)

    def test_search_no_perms(self):
        """Test search() as user with no permissions"""
        # Create user with no permissions to self.project
        user_no_perms = self.make_user('user_no_perms')
        self._add_event()
        ret = self.plugin.search(SEARCH_TERMS, user_no_perms)
        self.assertEqual(len(ret[0].items), 0)

    def test_search_mixed_perms(self):
        """Test search() as user with mixed permissions"""
        user_new = self.make_user('user_new')
        project_new = self.make_project(
            'TestProject2', PROJECT_TYPE_PROJECT, self.category
        )
        self.make_assignment(project_new, self.user_owner, self.role_owner)
        self.make_assignment(project_new, user_new, self.role_contributor)
        self._add_event()
        self.event_kw['project'] = project_new
        project_event_new = self._add_event()
        ret = self.plugin.search(SEARCH_TERMS, user_new)
        self.assertEqual(len(ret[0].items), 1)
        self.assertEqual(ret[0].items[0], project_event_new)
        ret = self.plugin.search(SEARCH_TERMS, self.user_owner)
        self.assertEqual(len(ret[0].items), 2)

    def test_search_project_classified_owner(self):
        """Test search() with classified project event as owner"""
        self.event_kw['classified'] = True
        self._add_event()
        ret = self.plugin.search(SEARCH_TERMS, self.user_owner)
        self.assertEqual(len(ret[0].items), 1)

    def test_search_project_classified_contributor(self):
        """Test search() with classified project event as contributor"""
        user_contrib = self.make_user('user_contrib')
        self.make_assignment(self.project, user_contrib, self.role_contributor)
        self.event_kw['classified'] = True
        self._add_event()
        ret = self.plugin.search(SEARCH_TERMS, user_contrib)
        self.assertEqual(len(ret[0].items), 0)

    def test_search_site_superuser(self):
        """Test search() with site event as superuser"""
        self.event_kw['project'] = None
        self._add_event()
        ret = self.plugin.search(SEARCH_TERMS, self.user)
        self.assertEqual(len(ret[0].items), 1)

    def test_search_site_regular_user(self):
        """Test search() with site event as regular user"""
        self.event_kw['project'] = None
        self._add_event()
        ret = self.plugin.search(SEARCH_TERMS, self.user_owner)
        self.assertEqual(len(ret[0].items), 1)

    def test_search_site_classified_superuser(self):
        """Test search() with classified site event as superuser"""
        self.event_kw['project'] = None
        self.event_kw['classified'] = True
        self._add_event()
        ret = self.plugin.search(SEARCH_TERMS, self.user)
        self.assertEqual(len(ret[0].items), 1)

    def test_search_site_classified_regular_user(self):
        """Test search() with classified site event as regular user"""
        self.event_kw['project'] = None
        self.event_kw['classified'] = True
        self._add_event()
        ret = self.plugin.search(SEARCH_TERMS, self.user_owner)
        self.assertEqual(len(ret[0].items), 0)

    def test_search_type(self):
        """Test search() with defined search type"""
        self._add_event()
        ret = self.plugin.search(
            SEARCH_TERMS, self.user, search_type='timeline'
        )
        self.assertEqual(len(ret[0].items), 1)

    def test_search_type_invalid(self):
        """Test search() with invalid search type"""
        self._add_event()
        ret = self.plugin.search(
            SEARCH_TERMS, self.user, search_type='raTho0Oo'
        )
        self.assertEqual(len(ret[0].items), 0)


class TestBackendPlugin(TimelinePluginTestBase):
    """Tests for timeline BackendPlugin"""

    def setUp(self):
        super().setUp()
        self.plugin = BackendPluginPoint.get_plugin(BACKEND_PLUGIN_NAME)

    def test_plugin_retrieval(self):
        """Test retrieving BackendPlugin"""
        self.assertIsNotNone(self.plugin)
        self.assertEqual(self.plugin.get_model().name, BACKEND_PLUGIN_NAME)
        self.assertEqual(self.plugin.name, BACKEND_PLUGIN_NAME)
        self.assertEqual(self.plugin.get_model().title, BACKEND_PLUGIN_TITLE)

    def test_get_api(self):
        """Test get_api()"""
        self.assertIsInstance(self.plugin.get_api(), TimelineAPI)


class TestSiteAppPlugin(TimelinePluginTestBase):
    """Tests for timeline SiteAppPlugin"""

    def setUp(self):
        super().setUp()
        self.plugin = SiteAppPluginPoint.get_plugin(SITE_PLUGIN_NAME)

    def test_plugin_retrieval(self):
        """Test retrieving SiteAppPlugin"""
        self.assertIsNotNone(self.plugin)
        self.assertEqual(self.plugin.get_model().name, SITE_PLUGIN_NAME)
        self.assertEqual(self.plugin.name, SITE_PLUGIN_NAME)
        self.assertEqual(self.plugin.get_model().title, SITE_PLUGIN_TITLE)
        self.assertEqual(self.plugin.urls, urls_ui_site)


class TestAdminSiteAppPlugin(TimelinePluginTestBase):
    """Tests for timeline AdminSiteAppPlugin"""

    def setUp(self):
        super().setUp()
        self.plugin = SiteAppPluginPoint.get_plugin(ADMIN_PLUGIN_NAME)

    def test_plugin_retrieval(self):
        """Test retrieving AdminSiteAppPlugin"""
        self.assertIsNotNone(self.plugin)
        self.assertEqual(self.plugin.get_model().name, ADMIN_PLUGIN_NAME)
        self.assertEqual(self.plugin.name, ADMIN_PLUGIN_NAME)
        self.assertEqual(self.plugin.get_model().title, ADMIN_PLUGIN_TITLE)
        self.assertEqual(self.plugin.urls, urls_ui_admin)



"""UI and Ajax view tests for the appalerts app"""

from django.urls import reverse

from test_plus.test import TestCase

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI

from appalerts.tests.test_models import AppAlertMixin


app_settings = AppSettingAPI()


# Local constants
APP_NAME_PR = 'projectroles'


class ViewTestBase(AppAlertMixin, TestCase):
    """Base class for appalerts view testing"""

    def setUp(self):
        # Create users
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        self.regular_user = self.make_user('regular_user')
        self.no_alert_user = self.make_user('no_alert_user')
        # No user
        self.anonymous = None
        # Create alert
        self.alert = self.make_app_alert(
            user=self.regular_user, url=reverse('home')
        )


class TestAppAlertListView(ViewTestBase):
    """Tests for AppAlertListView"""

    def setUp(self):
        super().setUp()
        self.url = reverse('appalerts:list')

    def test_get_superuser(self):
        """Test AppAlertListView GET as superuser"""
        with self.login(self.superuser):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['object_list'].count(), 0)
        self.assertEqual(response.context['read_only_disable'], False)

    def test_get_regular_user(self):
        """Test GET as user with assigned alert"""
        with self.login(self.regular_user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['object_list'].count(), 1)
        self.assertEqual(response.context['object_list'][0].pk, self.alert.pk)
        self.assertEqual(response.context['read_only_disable'], False)

    def test_get_no_alert_user(self):
        """Test GET as user without alerts"""
        with self.login(self.no_alert_user):
            response = self.client.get(self.url)
            self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['object_list'].count(), 0)
        self.assertEqual(response.context['read_only_disable'], False)

    def test_get_read_only_superuser(self):
        """Test GET with site read-only mode as superuser"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        with self.login(self.superuser):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['read_only_disable'], False)

    def test_get_read_only_regular_user(self):
        """Test GET with site read-only mode as regular user"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        with self.login(self.regular_user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['read_only_disable'], True)


class TestAppAlertRedirectView(ViewTestBase):
    """Tests for AppAlertLinkRedirectView"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'appalerts:redirect',
            kwargs={'appalert': self.alert.sodar_uuid},
        )
        self.list_url = reverse('appalerts:list')

    def test_get_superuser(self):
        """Test AppAlertLinkRedirectView GET as superuser"""
        self.assertEqual(self.alert.active, True)
        with self.login(self.superuser):
            response = self.client.get(self.url)
            self.assertRedirects(response, self.list_url)
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.active, True)

    def test_get_regular_user(self):
        """Test GET as user with assigned alert"""
        self.assertEqual(self.alert.active, True)
        with self.login(self.regular_user):
            response = self.client.get(self.url)
            self.assertRedirects(response, self.alert.url)
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.active, False)

    def test_get_no_alert_user(self):
        """Test GET as user without alerts"""
        self.assertEqual(self.alert.active, True)
        with self.login(self.no_alert_user):
            response = self.client.get(self.url)
            self.assertRedirects(response, self.list_url)
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.active, True)


class TestAppAlertStatusAjaxView(ViewTestBase):
    """Tests for AppAlertStatusAjaxView"""

    def test_get_user_with_alerts(self):
        """Test AppAlertStatusAjaxView GET as user with alert assigned"""
        with self.login(self.regular_user):
            response = self.client.get(reverse('appalerts:ajax_status'))
            self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['alerts'], 1)

    def test_get_user_no_alerts(self):
        """Test GET as user without alert assigned"""
        with self.login(self.no_alert_user):
            response = self.client.get(reverse('appalerts:ajax_status'))
            self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['alerts'], 0)


class TestAppAlertDismissAjaxView(ViewTestBase):
    """Tests for AppAlertDismissAjaxView"""

    def test_post_superuser(self):
        """Test AppAlertDismissAjaxView POST as superuser"""
        self.assertEqual(self.alert.active, True)
        with self.login(self.superuser):
            response = self.client.post(
                reverse(
                    'appalerts:ajax_dismiss',
                    kwargs={'appalert': self.alert.sodar_uuid},
                )
            )
            self.assertEqual(response.status_code, 404)
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.active, True)

    def test_post_regular_user(self):
        """Test POST as user with assigned alert"""
        self.assertEqual(self.alert.active, True)
        with self.login(self.regular_user):
            response = self.client.post(
                reverse(
                    'appalerts:ajax_dismiss',
                    kwargs={'appalert': self.alert.sodar_uuid},
                )
            )
            self.assertEqual(response.status_code, 200)
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.active, False)

    def test_post_no_alert_user(self):
        """Test POST as user without alerts"""
        self.assertEqual(self.alert.active, True)
        with self.login(self.no_alert_user):
            response = self.client.post(
                reverse(
                    'appalerts:ajax_dismiss',
                    kwargs={'appalert': self.alert.sodar_uuid},
                )
            )
            self.assertEqual(response.status_code, 404)
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.active, True)

    def test_post_regular_user_all(self):
        """Test POST as user dismissing all alerts"""
        self.assertEqual(self.alert.active, True)
        with self.login(self.regular_user):
            response = self.client.post(reverse('appalerts:ajax_dismiss_all'))
            self.assertEqual(response.status_code, 200)
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.active, False)

    def test_post_no_alert_user_all(self):
        """Test POST as user without alerts trying to dismiss all"""
        self.assertEqual(self.alert.active, True)
        with self.login(self.no_alert_user):
            response = self.client.post(reverse('appalerts:ajax_dismiss_all'))
            self.assertEqual(response.status_code, 404)
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.active, True)



import rules

# Projectroles dependency
from projectroles import rules as pr_rules  # To access common predicates


# Predicates -------------------------------------------------------------


# If we need to assign new predicates, we do it here


# Rules ------------------------------------------------------------------


# Rules should not be needed, use permissions for user rights


# Permissions ------------------------------------------------------------


# Allow viewing data in project
rules.add_perm(
    'filesfolders.view_data',
    pr_rules.is_project_owner
    | pr_rules.is_project_delegate
    | pr_rules.is_project_contributor
    | pr_rules.is_project_guest,
)

# Allow adding data to project
rules.add_perm(
    'filesfolders.add_data',
    pr_rules.can_modify_project_data
    & (
        pr_rules.is_project_owner
        | pr_rules.is_project_delegate
        | pr_rules.is_project_contributor
    ),
)

# Allow updating own data in project
rules.add_perm(
    'filesfolders.update_data_own',
    pr_rules.can_modify_project_data
    & (
        pr_rules.is_project_owner
        | pr_rules.is_project_delegate
        | pr_rules.is_project_contributor
    ),
)

# Allow updating all data in project
rules.add_perm(
    'filesfolders.update_data_all',
    pr_rules.can_modify_project_data
    & (pr_rules.is_project_owner | pr_rules.is_project_delegate),
)

# Allow sharing public temporary URLs
rules.add_perm(
    'filesfolders.share_public_link',
    pr_rules.is_project_owner
    | pr_rules.is_project_delegate
    | pr_rules.is_project_contributor,
)



"""REST API views for the timeline app"""

from rest_framework.exceptions import NotFound, PermissionDenied
from rest_framework.generics import ListAPIView, RetrieveAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.renderers import JSONRenderer
from rest_framework.versioning import AcceptHeaderVersioning

# Projectroles dependency
from projectroles.views_api import (
    SODARAPIGenericProjectMixin,
    SODARPageNumberPagination,
)

from timeline.models import TimelineEvent
from timeline.serializers import TimelineEventSerializer


# Local constants
TIMELINE_API_MEDIA_TYPE = 'application/vnd.bihealth.sodar-core.timeline+json'
TIMELINE_API_DEFAULT_VERSION = '2.0'
TIMELINE_API_ALLOWED_VERSIONS = ['2.0']


class TimelineAPIVersioningMixin:
    """
    Timeline API view versioning mixin for overriding media type and
    accepted versions.
    """

    class TimelineAPIRenderer(JSONRenderer):
        media_type = TIMELINE_API_MEDIA_TYPE

    class TimelineAPIVersioning(AcceptHeaderVersioning):
        allowed_versions = TIMELINE_API_ALLOWED_VERSIONS
        default_version = TIMELINE_API_DEFAULT_VERSION

    renderer_classes = [TimelineAPIRenderer]
    versioning_class = TimelineAPIVersioning


class ProjectTimelineEventListAPIView(
    TimelineAPIVersioningMixin, SODARAPIGenericProjectMixin, ListAPIView
):
    """
    List ``TimelineEvent`` objects belonging in a category or project. Events
    are ordered from newest to oldest.

    Supports optional pagination by providing the ``page`` query string. This
    will return results in the Django Rest Framework PageNumberPagination
    format.

    **URL:** ``/timeline/api/list/{Project.sodar_uuid}``

    **Methods:** ``GET``

    **Parameters:**

    - ``page``: Page number for paginated results (int, optional)

    **Returns:** List or paginated dict of ``TimelineEvent`` objects (see ``TimelineEventRetrieveAPIView``)
    """

    pagination_class = SODARPageNumberPagination
    permission_required = 'timeline.view_timeline'
    serializer_class = TimelineEventSerializer

    def get_queryset(self):
        project = self.get_project()
        user = self.request.user
        q_kwargs = {'project': project}
        if not user.is_superuser and not project.is_owner_or_delegate(user):
            q_kwargs['classified'] = False
        return TimelineEvent.objects.filter(**q_kwargs).order_by('-pk')


class SiteTimelineEventListAPIView(TimelineAPIVersioningMixin, ListAPIView):
    """
    List site-wide ``TimelineEvent`` objects. Events are ordered from newest to
    oldest.

    Supports optional pagination by providing the ``page`` query string. This
    will return results in the Django Rest Framework ``PageNumberPagination``
    format.

    **URL:** ``/timeline/api/list/site``

    **Methods:** ``GET``

    **Parameters:**

    - ``page``: Page number for paginated results (int, optional)

    **Returns:** List or paginated dict of ``TimelineEvent`` objects (see ``TimelineEventRetrieveAPIView``)
    """

    pagination_class = SODARPageNumberPagination
    permission_classes = [IsAuthenticated]
    serializer_class = TimelineEventSerializer

    def get_queryset(self):
        user = self.request.user
        q_kwargs = {'project': None}
        if not user.is_superuser:
            q_kwargs['classified'] = False
        return TimelineEvent.objects.filter(**q_kwargs).order_by('-pk')


class TimelineEventRetrieveAPIView(TimelineAPIVersioningMixin, RetrieveAPIView):
    """
    Retrieve ``TimelineEvent`` object.

    Extra data is only returned for users with sufficient permissions
    (superusers, project owners and delegates).

    **URL:** ``/timeline/api/retrieve/{TimelineEvent.sodar_uuid}``

    **Methods:** ``GET``

    **Returns:**

    - ``project``: Project UUID (string or None)
    - ``app``: App name (string)
    - ``user``: UUID of user who created the event (string or None)
    - ``event_name``: Event name (string)
    - ``description``: Event description (string)
    - ``extra_data``: Event extra data (JSON or None)
    - ``classified``: Whether event is classified (boolean)
    - ``status_changes``: List of TimelineEventStatus objects (JSON)
        - ``event``: TimelineEvent UUID (string)
        - ``timestamp``: Status datetime (YYYY-MM-DDThh:mm:ssZ)
        - ``status_type``: Status type (string)
        - ``description``: Status description (string)
        - ``extra_data``: Status extra data (JSON or None)
        - ``sodar_uuid``: Status UUID (string)
    - ``event_objects``: List of TimelineEventObjectRef objects (JSON)
        - ``event``: TimelineEvent UUID (string)
        - ``label``: Object label as given in event description (string)
        - ``name``: Name for identifying the object (string)
        - ``object_uuid``: UUID of the referred object (string)
        - ``extra_data``: Object reference extra data (JSON or None)
        - ``sodar_uuid``: Object reference UUID (string)
    - ``sodar_uuid``: TimelineEvent UUID (string)

    **Version Changes:**

    - ``2.0``: Return ``user`` as UUID instead of ``SODARUserSerializer`` dict
    """

    permission_classes = [IsAuthenticated]  # Perms checked in get_object()
    serializer_class = TimelineEventSerializer

    def get_object(self):
        obj = TimelineEvent.objects.filter(
            sodar_uuid=self.kwargs.get('timelineevent')
        ).first()
        if not obj:
            raise NotFound()
        user = self.request.user
        if obj.project and not user.has_perm(
            'timeline.view_timeline', obj.project
        ):
            raise PermissionDenied()
        if obj.classified:
            if (
                obj.project
                and not user.is_superuser
                and not obj.project.is_owner_or_delegate(user)
            ):
                raise PermissionDenied()
            elif not obj.project and not user.is_superuser:
                raise PermissionDenied()
        return obj



# Generated by Django 4.2.22 on 2025-06-19 10:44

from django.db import migrations


PROJECT_TYPE_PROJECT = 'PROJECT'
ROLE_NAME_GUEST = 'project guest'


def populate_public_access(apps, schema_editor):
    """Populate Project.public_access with former public_guest access values"""
    Project = apps.get_model('projectroles', 'Project')
    Role = apps.get_model('projectroles', 'Role')
    role_guest = Role.objects.get(name=ROLE_NAME_GUEST)
    for p in Project.objects.filter(type=PROJECT_TYPE_PROJECT):
        if p.public_guest_access:
            p.public_access = role_guest
            p.save()


def revert_public_access(apps, schema_editor):
    """Revert Project.public_access populating"""
    Project = apps.get_model('projectroles', 'Project')
    for p in Project.objects.filter(type=PROJECT_TYPE_PROJECT):
        if p.public_access:
            p.public_access = None
            p.save()


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0036_project_public_access'),
    ]
    operations = [
        migrations.RunPython(
            code=populate_public_access,
            reverse_code=revert_public_access,
        ),
    ]



import logging
from urllib.parse import ParseResult

from django.conf import settings
from django.contrib import auth
from django.views.generic import TemplateView

# Projectroles dependency
from projectroles.models import (
    Project,
    RemoteSite,
    SODAR_CONSTANTS,
    AUTH_TYPE_LDAP,
    AUTH_TYPE_LOCAL,
    AUTH_TYPE_OIDC,
)
from projectroles.plugins import PluginAPI
from projectroles.views import LoggedInPermissionMixin


logger = logging.getLogger(__name__)
plugin_api = PluginAPI()
User = auth.get_user_model()


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SYSTEM_USER_GROUP = SODAR_CONSTANTS['SYSTEM_USER_GROUP']

# Local constants
CORE_SETTINGS = [
    'ADMINS',
    'ALLOWED_HOSTS',
    'APPS_DIR',
    'AUTH_LDAP_CA_CERT_FILE',
    'AUTH_LDAP_DOMAIN_PRINTABLE',
    'AUTH_LDAP_SERVER_URI',
    'AUTH_LDAP_START_TLS',
    'AUTH_LDAP_USER_SEARCH_BASE',
    'AUTH_LDAP_USERNAME_DOMAIN',
    'AUTH_LDAP2_CA_CERT_FILE',
    'AUTH_LDAP2_DOMAIN_PRINTABLE',
    'AUTH_LDAP2_SERVER_URI',
    'AUTH_LDAP2_START_TLS',
    'AUTH_LDAP_USER_SEARCH_BASE',
    'AUTH_LDAP2_USERNAME_DOMAIN',
    'DEBUG',
    'EMAIL_BACKEND',
    'EMAIL_HOST',
    'EMAIL_PORT',
    'EMAIL_SENDER',
    'ENABLE_LDAP',
    'ENABLE_LDAP_SECONDARY',
    'ENABLE_OIDC',
    'ENABLED_BACKEND_PLUGINS',
    'ICONIFY_JSON_ROOT',
    'INSTALLED_APPS',
    'LANGUAGE_CODE',
    'LDAP_ALT_DOMAINS',
    'LDAP_DEBUG',
    'LOGGING_APPS',
    'LOGGING_FILE_PATH',
    'LOGGING_LEVEL',
    'MANAGERS',
    'MIDDLEWARE',
    'PROJECTROLES_ALLOW_ANONYMOUS',
    'PROJECTROLES_ALLOW_LOCAL_USERS',
    'PROJECTROLES_API_USER_DETAIL_RESTRICT',
    'PROJECTROLES_BROWSER_WARNING',
    'PROJECTROLES_CUSTOM_JS_INCLUDES',
    'PROJECTROLES_CUSTOM_CSS_INCLUDES',
    'PROJECTROLES_DEFAULT_ADMIN',
    'PROJECTROLES_DELEGATE_LIMIT',
    'PROJECTROLES_DISABLE_CATEGORIES',
    'PROJECTROLES_DISABLE_CDN_INCLUDES',
    'PROJECTROLES_EMAIL_SENDER_REPLY',
    'PROJECTROLES_ENABLE_PROFILING',
    'PROJECTROLES_ENABLE_SEARCH',
    'PROJECTROLES_HELP_HIGHLIGHT_DAYS',
    'PROJECTROLES_HIDE_PROJECT_APPS',
    'PROJECTROLES_INLINE_HEAD_INCLUDE',
    'PROJECTROLES_INVITE_EXPIRY_DAYS',
    'PROJECTROLES_KIOSK_MODE',
    'PROJECTROLES_SEARCH_PAGINATION',
    'PROJECTROLES_SECRET_LENGTH',
    'PROJECTROLES_SEND_EMAIL',
    'PROJECTROLES_SIDEBAR_ICON_SIZE',
    'PROJECTROLES_SITE_MODE',
    'PROJECTROLES_TEMPLATE_INCLUDE_PATH',
    'PROJECTROLES_TARGET_CREATE',
    'PROJECTROLES_TARGET_SYNC_ENABLE',
    'PROJECTROLES_TARGET_SYNC_INTERVAL',
    'ROOT_DIR',
    'SITE_PACKAGE',
    'SITE_TITLE',
    'SITE_SUBTITLE',
    'SITE_INSTANCE_TITLE',
    'SOCIAL_AUTH_OIDC_OIDC_ENDPOINT',
    'SOCIAL_AUTH_OIDC_KEY',
    'SOCIAL_AUTH_OIDC_USERNAME_KEY',
    'SODAR_API_DEFAULT_HOST',
    'SODAR_API_PAGE_SIZE',
    'STATICFILES_DIRS',
    'STATICFILES_FINDERS',
    'TEMPLATES',
    'TIME_ZONE',
    'USE_TZ',
]


class SiteInfoView(LoggedInPermissionMixin, TemplateView):
    """Site info view"""

    permission_required = 'siteinfo.view_info'
    template_name = 'siteinfo/site_info.html'

    @classmethod
    def _get_settings(cls, keys: list[str]) -> dict:
        ret = {}
        for k in keys:
            if hasattr(settings, k):
                v = getattr(settings, k)
                if isinstance(v, ParseResult):
                    v = v.geturl()
                ret[k] = {'value': v, 'set': True}
            else:
                ret[k] = {'set': False}
        return ret

    @classmethod
    def _get_plugin_stats(cls, p_list: list) -> dict:
        ret = {}
        for p in p_list:
            try:
                ret[p] = {'stats': p.get_statistics(), 'settings': {}}
            except Exception as ex:
                ret[p] = {'error': str(ex)}
                logger.error(f'Exception in {p.name}.get_statistics(): {ex}')
            if p.info_settings:
                try:
                    ret[p]['settings'] = cls._get_settings(p.info_settings)
                except Exception as ex:
                    logger.error(
                        f'Exception in _get_settings() for {p.name}: {ex}'
                    )
        return ret

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)

        # Project statistics
        context['project_count'] = Project.objects.filter(
            type=PROJECT_TYPE_PROJECT
        ).count()
        context['category_count'] = Project.objects.filter(
            type=PROJECT_TYPE_CATEGORY
        ).count()
        context['project_archive_count'] = Project.objects.filter(
            type=PROJECT_TYPE_PROJECT, archive=True
        ).count()

        # User statistics
        users = User.objects.all()
        context['user_total_count'] = users.count()
        context['user_ldap_count'] = len(
            [u for u in users if u.get_auth_type() == AUTH_TYPE_LDAP]
        )
        context['user_oidc_count'] = len(
            [u for u in users if u.get_auth_type() == AUTH_TYPE_OIDC]
        )
        local_users = [
            u
            for u in users
            if u.get_auth_type() == AUTH_TYPE_LOCAL and not u.is_superuser
        ]
        context['user_local_count'] = len(local_users)
        context['user_admin_count'] = User.objects.filter(
            is_superuser=True
        ).count()
        context['user_inactive_count'] = User.objects.filter(
            is_active=False
        ).count()

        # App plugins
        project_plugins = plugin_api.get_active_plugins('project_app')
        backend_plugins = plugin_api.get_active_plugins('backend')
        site_plugins = plugin_api.get_active_plugins('site_app')

        # Plugin statistics
        context['project_plugins'] = self._get_plugin_stats(project_plugins)
        context['site_plugins'] = self._get_plugin_stats(site_plugins)
        context['backend_plugins'] = self._get_plugin_stats(backend_plugins)

        # Basic site info
        context['site_title'] = settings.SITE_TITLE
        context['site_subtitle'] = settings.SITE_SUBTITLE
        context['site_instance_title'] = settings.SITE_INSTANCE_TITLE

        # Remote site info
        context['site_mode'] = settings.PROJECTROLES_SITE_MODE

        if settings.PROJECTROLES_SITE_MODE == SITE_MODE_SOURCE:
            context['site_target_count'] = RemoteSite.objects.filter(
                mode=SITE_MODE_TARGET
            ).count()

        # Core settings
        context['settings_core'] = self._get_settings(CORE_SETTINGS)
        # TODO: Add LDAP settings?
        return context



"""Template tags for the timeline app"""

import logging

from typing import Optional

from django import template
from django.db.models import QuerySet
from django.http import HttpRequest
from django.urls import reverse
from django.utils.timezone import localtime

from djangoplugins.models import Plugin

# Projectroles dependency
from projectroles.models import Project
from projectroles.plugins import ProjectAppPluginPoint

from timeline.api import TimelineAPI
from timeline.models import TimelineEvent


logger = logging.getLogger(__name__)
timeline = TimelineAPI()
register = template.Library()


ICON_PROJECTROLES = 'mdi:cube'
ICON_UNKNOWN_APP = 'mdi:help-circle'
STATUS_STYLES = {
    'OK': 'bg-success',
    'INIT': 'bg-secondary',
    'SUBMIT': 'bg-warning',
    'FAILED': 'bg-danger',
    'INFO': 'bg-info',
    'CANCEL': 'bg-dark',
}


# Event helpers ----------------------------------------------------------------


@register.simple_tag
def get_timestamp(event: TimelineEvent) -> str:
    """Return printable timestamp of event in local timezone"""
    try:
        return localtime(event.get_timestamp()).strftime('%Y-%m-%d %H:%M:%S')
    except Exception:  # Handle error cases where TimelineEventStatus is missing
        return 'N/A'


@register.simple_tag
def get_event_name(event: TimelineEvent) -> str:
    """Return timeline event name in a user friendly format"""
    return event.event_name.replace('_', ' ').title()


@register.simple_tag
def get_event_description(
    event: TimelineEvent,
    plugin_lookup: dict,
    request: Optional[HttpRequest] = None,
) -> str:
    """Return printable version of event description"""
    return timeline.get_event_description(event, plugin_lookup, request)


@register.simple_tag
def get_details_events(
    project: Project, view_classified: bool = False
) -> QuerySet:
    """Return recent events for card on project details page"""
    c_kwargs = {'classified': False} if not view_classified else {}
    return TimelineEvent.objects.filter(project=project, **c_kwargs).order_by(
        '-pk'
    )[:5]


@register.simple_tag
def get_plugin_lookup() -> dict:
    """Return lookup dict of app plugins with app name as key"""
    ret = {}
    for p in Plugin.objects.all():
        try:
            ret[p.name] = p.get_plugin()
        except Exception as ex:
            logger.error(f'Unable to retrieve plugin "{p.name}": {ex}')
    return ret


# Template rendering -----------------------------------------------------------


@register.simple_tag
def get_app_icon_html(event: TimelineEvent, plugin_lookup: dict) -> str:
    """Return icon link HTML for app by plugin lookup"""
    url = None
    title = event.app
    icon = ICON_UNKNOWN_APP  # Default in case the plugin is not found

    if event.app == 'projectroles':
        if event.project:
            url = reverse(
                'projectroles:detail',
                kwargs={'project': event.project.sodar_uuid},
            )
        title = 'Projectroles'
        icon = ICON_PROJECTROLES
    else:
        plugin_name = event.plugin if event.plugin else event.app
        if plugin_name in plugin_lookup.keys():
            url_kwargs = {}
            plugin = plugin_lookup[plugin_name]
            if isinstance(plugin, ProjectAppPluginPoint):
                url_kwargs['project'] = event.project.sodar_uuid
            entry_point = getattr(plugin, 'entry_point_url_id', None)
            if entry_point:
                try:
                    url = reverse(entry_point, kwargs=url_kwargs)
                except Exception as ex:
                    url = None
                    logger.error(
                        f'Unable to get URL for entry point "{entry_point}": '
                        f'{ex}'
                    )
            title = plugin.title
            if getattr(plugin, 'icon', None):
                icon = plugin.icon

    return (
        '<a {} class="sodar-tl-item-app-icon" title="{}" data-toggle="tooltip" '
        'data-placement="top">'
        '<i class="iconify" data-icon="{}"></i></a>'.format(
            'href="{}"'.format(url) if url else '', title, icon
        )
    )


@register.simple_tag
def get_status_style(status: str) -> str:
    """Retrn status style class"""
    return (
        (STATUS_STYLES[status.status_type] + ' text-light')
        if status.status_type in STATUS_STYLES
        else 'bg-light'
    )



"""Template tags provided by projectroles for use in other apps"""

import mistune

from importlib import import_module
from typing import Any, Optional, Union
from uuid import UUID

from django import template
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.staticfiles import finders
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpRequest
from django.template.loader import get_template
from django.templatetags.static import static
from django.urls import reverse

import projectroles
from projectroles.app_settings import AppSettingAPI
from projectroles.models import (
    Project,
    Role,
    RemoteProject,
    SODARUser,
    SODAR_CONSTANTS,
)
from projectroles.plugins import BackendPluginPoint, PluginAPI
from projectroles.utils import get_display_name as _get_display_name


app_settings = AppSettingAPI()
plugin_api = PluginAPI()
site = import_module(settings.SITE_PACKAGE)
User = get_user_model()

register = template.Library()


# SODAR constants
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
SITE_MODE_PEER = SODAR_CONSTANTS['SITE_MODE_PEER']


# SODAR and site operations ----------------------------------------------------


@register.simple_tag
def site_version() -> str:
    """Return the site version"""
    return site.__version__ if hasattr(site, '__version__') else '[UNKNOWN]'


@register.simple_tag
def core_version() -> str:
    """Return the SODAR Core version"""
    return projectroles.__version__


@register.simple_tag
def check_backend(name: str) -> bool:
    """Return True if backend app is available and enabled, else False"""
    return True if name in settings.ENABLED_BACKEND_PLUGINS else False


@register.simple_tag
def get_project_by_uuid(sodar_uuid: Union[str, UUID]) -> Optional[Project]:
    """Return Project by sodar_uuid"""
    return Project.objects.filter(sodar_uuid=sodar_uuid).first()


@register.simple_tag
def get_user_by_uuid(sodar_uuid: Union[str, UUID]) -> Optional[SODARUser]:
    """Return SODARUser by sodar_uuid"""
    return User.objects.filter(sodar_uuid=sodar_uuid).first()


@register.simple_tag
def get_user_by_username(username: str) -> Optional[SODARUser]:
    """Return User by username"""
    return User.objects.filter(username=username).first()


# Django helpers ---------------------------------------------------------------


@register.simple_tag
def get_django_setting(name: str, default: Any = None, js: bool = False) -> Any:
    """
    Return value of Django setting by name or the default value if the setting
    is not found. Return a Javascript-safe value if js=True.
    """
    val = getattr(settings, name, default)
    if js and isinstance(val, bool):
        val = int(val)
    return val


@register.simple_tag
def get_app_setting(
    plugin_name: str,
    setting_name: str,
    project: Optional[Project] = None,
    user: Optional[SODARUser] = None,
) -> Any:
    """Get a project/user specific app setting value from AppSettingAPI"""
    return app_settings.get(plugin_name, setting_name, project, user)


@register.simple_tag
def static_file_exists(path: str) -> bool:
    """Return True/False based on whether a static file exists"""
    return True if finders.find(path) else False


@register.simple_tag
def template_exists(path: str) -> bool:
    """Return True/False based on whether a template exists"""
    try:
        get_template(path)
        return True
    except template.TemplateDoesNotExist:
        return False


@register.simple_tag
def get_full_url(request: HttpRequest, url: str) -> str:
    """Get full URL based on a local URL"""
    return request.scheme + '://' + request.get_host() + url


# Template rendering -----------------------------------------------------------


@register.simple_tag
def get_display_name(
    key: str, title: bool = False, count: int = 1, plural: bool = False
) -> str:
    """Return display name from SODAR_CONSTANTS"""
    return _get_display_name(key, title, count, plural)


@register.simple_tag
def get_role_display_name(
    role: Role, project: Project, title: bool = False
) -> str:
    """
    Return display name for role assignment.

    :param role: Role object
    :param project: Project object for context (to support inheritance)
    :param title: Return in title case if True (boolean)
    """
    role_suffix = role.name.split(' ')[1]
    if title:
        role_suffix = role_suffix.title()
    return f'{_get_display_name(project.type, title=title)} {role_suffix}'


@register.simple_tag
def get_project_title_html(project: Project) -> str:
    """Return HTML version of the full project title including parents"""
    ret = ''
    if project.get_parents():
        ret += ' / '.join(project.full_title.split(' / ')[:-1]) + ' / '
    ret += project.title
    return ret


@register.simple_tag
def get_project_link(
    project: Project,
    full_title: bool = False,
    request: Optional[HttpRequest] = None,
) -> str:
    """Return link to project with a simple or full title"""
    remote_icon = ''
    if request:
        remote_icon = get_remote_icon(project, request)
    return (
        '<a href="{}" title="{}" data-toggle="tooltip" '
        'data-placement="top">{}</a>{}'.format(
            reverse(
                'projectroles:detail', kwargs={'project': project.sodar_uuid}
            ),
            project.description if project.description else '',
            project.full_title if full_title else project.title,
            ' ' + remote_icon if remote_icon else '',
        )
    )


@register.simple_tag
def get_user_superuser_icon(tooltip: bool = True) -> str:
    """Return superuser icon for user"""
    ret = ''
    if tooltip:
        ret += '<span title="Superuser" data-toggle="tooltip">'
    ret += (
        '<i class="iconify text-info ml-1" '
        'data-icon="mdi:shield-account"></i>'
    )
    if tooltip:
        ret += '</span>'
    return ret


@register.simple_tag
def get_user_inactive_icon(tooltip: bool = True) -> str:
    """Return inactive icon for user"""
    ret = ''
    if tooltip:
        ret += '<span title="Inactive" data-toggle="tooltip">'
    ret += (
        '<i class="iconify text-secondary ml-1" '
        'data-icon="mdi:account-off"></i>'
    )
    if tooltip:
        ret += '</span>'
    return ret


@register.simple_tag
def get_user_html(user: SODARUser) -> str:
    """Return HTML representation of a User object"""
    email_link = True if user.is_active and user.email else False
    title = user.get_full_name()
    if not user.is_active:
        title += ' (inactive)'
    elif user.is_superuser:
        title += ' (superuser)'
    ret = (
        '<span class="sodar-user-html{}" data-toggle="tooltip" '
        'data-placement="top" title="{}">'.format(
            ' text-secondary' if not user.is_active else '', title
        )
    )
    if email_link:
        ret += f'<a href="mailto:{user.email}">'
    ret += user.username
    if email_link:
        ret += '</a>'
    if user.is_superuser:
        ret += get_user_superuser_icon(False)
    if not user.is_active:
        ret += get_user_inactive_icon(False)
    ret += '</span>'
    return ret


@register.simple_tag
def get_user_badge(user: SODARUser, extra_class: Optional[str] = None) -> str:
    """
    Return badge HTML for a User object.

    :param user: SODARUser object
    :param extra_class: Extra classes to add to element (string, optional)
    :return: String (contains HTML)
    """
    if not user.is_active:
        icon = 'mdi:account-off'
        variant = 'secondary'
        user_class = 'inactive'
    elif user.is_superuser:
        icon = 'mdi:shield-account'
        variant = 'primary'
        user_class = 'superuser'
    else:
        icon = 'mdi:account'
        variant = 'primary'
        user_class = 'active'
    email = True if user.is_active and user.email else False
    if extra_class and not extra_class.startswith(' '):
        extra_class = ' ' + extra_class
    elif not extra_class:
        extra_class = ''
    ret = (
        f'<span class="badge badge-{variant} sodar-obj-badge '
        f'sodar-user-badge sodar-user-badge-{user_class}{extra_class}" '
        f'title="{user.get_full_name()}" data-toggle="tooltip" '
        f'data-uuid="{user.sodar_uuid}">'
        f'<i class="iconify" data-icon="{icon}"></i> '
    )
    if email:
        ret += f'<a class="text-white" href="mailto:{user.email}">'
    ret += user.username
    if email:
        ret += '</a>'
    ret += '</span>'
    return ret


@register.simple_tag
def get_project_badge(
    project: Project,
    variant: Optional[str] = None,
    can_view: bool = True,
    extra_class: Optional[str] = None,
) -> str:
    """
    Return badge HTML for a Project object.

    :param project: Project object
    :param variant: Badge variant (string)
    :param can_view: Display project link if True (bool)
    :param extra_class: Extra classes to add to element (string, optional)
    :return: String (contains HTML)
    """
    if not can_view:
        variant = 'secondary'
    elif not variant:
        variant = 'info'
    extra_class = (' ' + extra_class) if extra_class else ''
    ret = (
        f'<span class="badge badge-{variant.lower()} sodar-obj-badge '
        f'sodar-project-badge{extra_class}" title="{project.full_title}" '
        f'data-toggle="tooltip">'
    )
    if project.is_project():
        icon = 'mdi:cube'
    else:
        icon = 'mdi:rhombus-split'
    ret += f'<i class="iconify" data-icon="{icon}"></i> '
    if can_view:
        url = reverse(
            'projectroles:detail', kwargs={'project': project.sodar_uuid}
        )
        ret += f'<a href="{url}">{project.title}</a>'
    else:
        ret += project.title
    ret += '</span>'
    return ret


@register.simple_tag
def get_backend_include(backend_name: str, include_type: str = 'js') -> str:
    """
    Return import string for backend app Javascript or CSS. Returns empty string
    if not found.
    """
    # TODO: Replace with get_app_plugin() and if None check
    # TODO: once get_app_plugin() can be used for backend plugins
    # TODO: Don't forget to remove ObjectDoesNotExist import
    try:
        plugin = BackendPluginPoint.get_plugin(backend_name)
    except ObjectDoesNotExist:
        return ''

    include = ''
    include_string = ''
    try:
        if include_type == 'js':
            include = plugin.javascript_url
            include_string = '<script type="text/javascript" src="{}"></script>'
        elif include_type == 'css':
            include = plugin.css_url
            include_string = (
                '<link rel="stylesheet" type="text/css" href="{}"/>'
            )
    except AttributeError:
        return ''

    if include and finders.find(include):
        return include_string.format(static(include))

    return ''


@register.simple_tag
def get_history_dropdown(obj: Any, project: Optional[Project] = None) -> str:
    """Return link to object timeline events within project"""
    timeline = plugin_api.get_backend_api('timeline_backend')
    if not timeline:
        return ''
    url = timeline.get_object_url(obj, project)
    return (
        '<a class="dropdown-item sodar-pr-role-link-history" href="{}">\n'
        '<i class="iconify" data-icon="mdi:clock-time-eight-outline"></i> '
        'History</a>\n'.format(url)
    )


@register.simple_tag
def highlight_search_term(item: str, terms: list[str]) -> str:
    """Return string with search term highlighted"""
    # Skip highlighting for multiple terms (at least for now)
    if isinstance(terms, list) and len(terms) > 1:
        return item
    elif isinstance(terms, list) and len(terms) == 1:
        term = terms[0]
    else:
        term = terms  # Old implementation
    if not term:  # If something goes wrong and we end up with no search term
        return item

    def get_highlights(item):
        pos = item.lower().find(term.lower())
        tl = len(term)
        if pos == -1:
            return item  # Nothing to highlight
        ret = item[:pos]
        ret += (
            '<span class="sodar-search-highlight">'
            + item[pos : pos + tl]
            + '</span>'
        )
        if len(item[pos + tl :]) > 0:
            ret += get_highlights(item[pos + tl :])
        return ret

    return get_highlights(item)


@register.simple_tag
def get_info_link(content: str, html: bool = False) -> str:
    """Return info popover link icon"""
    return (
        '<a class="sodar-info-link" tabindex="0" data-toggle="popover" '
        'data-trigger="focus" data-placement="top" data-content="{}" {}>'
        '<i class="iconify text-info" data-icon="mdi:information"></i>'
        '</a>'.format(content, 'data-html="true"' if html else '')
    )


@register.simple_tag
def get_remote_icon(project: Project, request: HttpRequest) -> str:
    """Get remote project icon HTML"""
    if project.is_remote() and request.user.is_superuser:
        remote_project = RemoteProject.objects.filter(
            project=project, site__mode=SITE_MODE_SOURCE
        ).first()
        if remote_project:
            return (
                '<i class="iconify {} mx-1 '
                'sodar-pr-remote-project-icon" data-icon="mdi:cloud" '
                'title="{} project from '
                '{}" data-toggle="tooltip" data-placement="top">'
                '</i>'.format(
                    'text-danger' if project.is_revoked() else 'text-info',
                    'REVOKED remote' if project.is_revoked() else 'Remote',
                    remote_project.site.name,
                )
            )
    return ''


@register.simple_tag
def render_markdown(raw_markdown: str) -> str:
    """Markdown field rendering helper"""
    return mistune.html(raw_markdown)


@register.filter
def force_wrap(s: str, length: int) -> str:
    """Force wrapping of string"""
    # If string contains spaces or hyphens, leave wrapping to browser
    if not {' ', '-'}.intersection(s) and len(s) > length:
        return '<wbr />'.join(
            [s[i : i + length] for i in range(0, len(s), length)]
        )
    return s


# General helpers  -------------------------------------------------------------


@register.simple_tag
def get_class(obj: Any, lower: bool = False) -> str:
    """Return object class as string"""
    c = obj.__class__.__name__
    return c.lower() if lower else c


@register.filter
def split(s: str, sep: str) -> list:
    """Split string by separator"""
    return s.split(sep)



"""Celery tasks for the samplesheets app"""

import logging

from django.conf import settings

from config.celery import app

from projectroles.app_settings import AppSettingAPI
from projectroles.remote_projects import RemoteProjectAPI
from projectroles.models import RemoteSite, SODAR_CONSTANTS


app_settings = AppSettingAPI()
logger = logging.getLogger(__name__)


# SODAR constants
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']

# Local constants
APP_NAME = 'projectroles'
READ_ONLY_SKIP_MSG = 'Site read-only mode enabled, skipping'


@app.task(bind=True)
def sync_remote_site_task(_self):
    """Synchronize remote project"""
    remote_api = RemoteProjectAPI()
    source_site = RemoteSite.objects.filter(mode=SITE_MODE_SOURCE).first()
    if source_site:
        if app_settings.get(APP_NAME, 'site_read_only'):
            logger.info(READ_ONLY_SKIP_MSG)
            return
        try:
            remote_data = remote_api.get_remote_data(source_site)
            logger.debug(remote_data)
            remote_api.sync_remote_data(source_site, remote_data)
        except Exception as ex:
            logger.error(f'Unable to synchronize {source_site.name}: {ex}')


@app.on_after_finalize.connect
def setup_periodic_tasks(sender, **kwargs):
    if settings.PROJECTROLES_SITE_MODE == SITE_MODE_TARGET and getattr(
        settings, 'PROJECTROLES_TARGET_SYNC_ENABLE', False
    ):
        sender.add_periodic_task(
            getattr(settings, 'PROJECTROLES_TARGET_SYNC_INTERVAL', 5) * 60,
            sync_remote_site_task.s(),
            name='synchronize_remote_site',
        )



"""Syncremote management command for synchronizing remote projects"""

import sys

from django.conf import settings
from django.core.management.base import BaseCommand

from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import RemoteSite, SODAR_CONSTANTS
from projectroles.remote_projects import RemoteProjectAPI


logger = ManagementCommandLogger(__name__)


# SODAR constants
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']


class Command(BaseCommand):
    help = 'Synchronizes user and project data from the source site'

    def add_arguments(self, parser):
        pass

    def handle(self, *args, **options):
        remote_api = RemoteProjectAPI()
        if getattr(settings, 'PROJECTROLES_DISABLE_CATEGORIES', False):
            logger.info(
                'Project categories and nesting disabled, '
                'remote sync disabled'
            )
            sys.exit(0)
        if settings.PROJECTROLES_SITE_MODE != SITE_MODE_TARGET:
            logger.error('Site not in TARGET mode, unable to sync')
            sys.exit(1)
        try:
            source_site = RemoteSite.objects.get(mode=SITE_MODE_SOURCE)
        except RemoteSite.DoesNotExist:
            logger.error('No source site set, unable to sync')
            sys.exit(1)

        if getattr(settings, 'PROJECTROLES_ALLOW_LOCAL_USERS', False):
            logger.info(
                'PROJECTROLES_ALLOW_LOCAL_USERS=True, will sync '
                'roles for existing local users'
            )
        logger.info(
            f'Retrieving data from source site "{source_site.name}" '
            f'({source_site.get_url()})..'
        )
        try:
            remote_data = remote_api.get_remote_data(source_site)
        except Exception as ex:
            logger.error(f'Failed to retrieve data from source site: {ex}')
            sys.exit(1)

        logger.info('Synchronizing remote data from source site..')
        try:
            remote_api.sync_remote_data(source_site, remote_data)
        except Exception as ex:
            logger.error(f'Remote sync failed with exception: {ex}')
            sys.exit(1)
        logger.info('Syncremote command OK')



"""Plugins for the adminalerts app"""

from typing import Optional

from django.urls import reverse
from django.utils import timezone

# Projectroles dependency
from projectroles.models import SODARUser, SODAR_CONSTANTS
from projectroles.plugins import SiteAppPluginPoint, PluginAppSettingDef

from adminalerts.models import AdminAlert
from adminalerts.urls import urlpatterns


# SODAR constants
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']

# Local constants
ADMINALERTS_APP_SETTINGS = [
    PluginAppSettingDef(
        name='notify_email_alert',
        scope=APP_SETTING_SCOPE_USER,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=True,
        label='Receive email for admin alerts',
        description=(
            'Receive email for important administrator alerts regarding e.g. '
            'site downtime.'
        ),
    )
]


class SiteAppPlugin(SiteAppPluginPoint):
    """Projectroles plugin for registering the app"""

    #: Name (used as plugin ID)
    name = 'adminalerts'

    #: Title (used in templates)
    title = 'Admin Alerts'

    #: UI URLs
    urls = urlpatterns

    #: App setting definitions
    app_settings = ADMINALERTS_APP_SETTINGS

    #: Iconify icon
    icon = 'mdi:alert'

    #: Description string
    description = 'Administrator alerts to be shown for users'

    #: Entry point URL ID
    entry_point_url_id = 'adminalerts:list'

    #: Required permission for displaying the app
    app_permission = 'adminalerts.create_alert'

    #: Names of plugin specific Django settings to display in siteinfo
    info_settings = ['ADMINALERTS_PAGINATION']

    def get_statistics(self) -> dict:
        return {
            'alert_count': {
                'label': 'Alerts',
                'value': AdminAlert.objects.all().count(),
            }
        }

    def get_messages(self, user: Optional[SODARUser] = None) -> list[dict]:
        """
        Return a list of messages to be shown to users.

        :param user: User object (optional)
        :return: List of dicts or empty list if no messages
        """
        messages = []
        alerts = AdminAlert.objects.filter(
            active=True, date_expire__gte=timezone.now()
        ).order_by('-pk')

        for a in alerts:
            content = '<i class="iconify" data-icon="mdi:alert"></i> '
            if a.description.raw and user and user.is_authenticated:
                url = reverse(
                    'adminalerts:detail',
                    kwargs={'adminalert': a.sodar_uuid},
                )
                content += (
                    f'<a href="{url}" '
                    f'class="sodar-alert-full-text-link">{a.message}</a> '
                )
                content += (
                    f'<span class="pull-right"><a href="{url}" '
                    f'class="text-info sodar-alert-detail-link">'
                    f'<i class="iconify" data-icon="mdi:arrow-right-circle">'
                    f'</i> Details</a>'
                )
            else:
                content += a.message
            messages.append(
                {
                    'content': content,
                    'color': 'info',
                    'dismissable': False,
                    'require_auth': a.require_auth,
                }
            )
        return messages



"""Ajax API view tests for the projectroles app"""

import json

from django.test import override_settings
from django.urls import reverse
from django.utils import timezone

from test_plus.test import TestCase

from projectroles.app_links import AppLinkAPI
from projectroles.app_settings import AppSettingAPI
from projectroles.models import AppSetting, SODAR_CONSTANTS
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleAssignmentMixin,
    RemoteSiteMixin,
    RemoteProjectMixin,
)
from projectroles.tests.test_views import (
    ViewTestBase,
    PROJECT_TYPE_CATEGORY,
    PROJECT_TYPE_PROJECT,
)
from projectroles.tests.test_views_api import SerializedObjectMixin


app_links = AppLinkAPI()
app_settings = AppSettingAPI()


# SODAR constants
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
REMOTE_LEVEL_READ_ROLES = SODAR_CONSTANTS['REMOTE_LEVEL_READ_ROLES']

# Local constants
APP_NAME = 'projectroles'
APP_NAME_FF = 'filesfolders'
INVALID_UUID = '11111111-1111-1111-1111-111111111111'


class TestProjectListAjaxView(ProjectMixin, RoleAssignmentMixin, ViewTestBase):
    """Tests for ProjectListAjaxView"""

    def setUp(self):
        super().setUp()
        self.user_owner_cat = self.make_user('user_owner_cat')
        self.user_contributor_cat = self.make_user('user_contributor_cat')
        self.user_owner = self.make_user('user_owner')
        self.user_no_roles = self.make_user('user_no_roles')
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user_owner_cat, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        self.cat_contributor_as = self.make_assignment(
            self.project, self.user_contributor_cat, self.role_contributor
        )
        self.url = reverse('projectroles:ajax_project_list')

    def test_get(self):
        """Test TestProjectListAjaxView GET"""
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        expected = {
            'projects': [
                {
                    'type': self.category.type,
                    'full_title': self.category.full_title,
                    'public_access': False,
                    'public_stats': False,
                    'archive': False,
                    'remote': False,
                    'revoked': False,
                    'starred': False,
                    'access': True,
                    'finder_url': None,
                    'uuid': str(self.category.sodar_uuid),
                },
                {
                    'type': self.project.type,
                    'full_title': self.project.full_title,
                    'public_access': False,
                    'public_stats': False,
                    'archive': False,
                    'remote': False,
                    'revoked': False,
                    'blocked': False,
                    'starred': False,
                    'access': True,
                    'finder_url': None,
                    'uuid': str(self.project.sodar_uuid),
                },
            ],
            'parent_depth': 0,
            'messages': {},
            'user': {'highlight': True, 'superuser': True},
        }
        self.assertEqual(response.data, expected)

    def test_get_parent(self):
        """Test GET with parent project"""
        with self.login(self.user):
            response = self.client.get(
                self.url + '?parent=' + str(self.category.sodar_uuid),
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            'projects': [
                {
                    'type': self.project.type,
                    'full_title': self.project.title,  # Not full_title
                    'public_access': False,
                    'public_stats': False,
                    'archive': False,
                    'remote': False,
                    'revoked': False,
                    'blocked': False,
                    'starred': False,
                    'access': True,
                    'finder_url': None,
                    'uuid': str(self.project.sodar_uuid),
                },
            ],
            'parent_depth': 1,
            'messages': {},
            'user': {'highlight': True, 'superuser': True},
        }
        self.assertEqual(response.data, expected)

    def test_get_highlight(self):
        """Test GET with highlight app setting enabled"""
        app_settings.set(
            APP_NAME, 'project_list_highlight', True, user=self.user
        )
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['user']['highlight'], True)

    def test_get_inherited_owner(self):
        """Test GET as inherited owner"""
        with self.login(self.user_owner_cat):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['projects']), 2)

    def test_get_inherited_contrib(self):
        """Test GET as inherited contributor"""
        with self.login(self.user_contributor_cat):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['projects']), 2)

    def test_get_no_roles(self):
        """Test project list as user with no roles"""
        with self.login(self.user_no_roles):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['projects'], [])
        self.assertIsNotNone(response.data['messages'].get('no_projects'))

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET as anonymous user"""
        # with self.login(self.user_no_roles):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['projects'], [])
        self.assertIsNotNone(response.data['messages'].get('no_projects'))

    def test_get_project_parent(self):
        """Test project list with project as parent (should fail)"""
        with self.login(self.user):
            response = self.client.get(
                self.url + '?parent=' + str(self.project.sodar_uuid),
            )
        self.assertEqual(response.status_code, 400)

    def test_get_block_owner(self):
        """Test GET with project access block as owner"""
        app_settings.set(
            APP_NAME, 'project_access_block', True, project=self.project
        )
        with self.login(self.user_owner):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertNotIn('blocked', pd[0])
        self.assertEqual(pd[1]['blocked'], True)
        self.assertEqual(pd[1]['access'], False)

    def test_get_block_superuser(self):
        """Test GET with project access block as superuser"""
        app_settings.set(
            APP_NAME, 'project_access_block', True, project=self.project
        )
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertNotIn('blocked', pd[0])
        self.assertEqual(pd[1]['blocked'], True)
        self.assertEqual(pd[1]['access'], True)

    def test_get_finder(self):
        """Test GET with finder role"""
        user_finder = self.make_user('user_finder')
        self.make_assignment(self.category, user_finder, self.role_finder)
        with self.login(user_finder):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertEqual(len(pd), 2)
        self.assertEqual(pd[0]['full_title'], self.category.full_title)
        self.assertEqual(pd[0]['access'], True)
        self.assertEqual(pd[0]['finder_url'], None)
        self.assertEqual(pd[1]['full_title'], self.project.full_title)
        self.assertEqual(pd[1]['access'], False)
        self.assertEqual(
            pd[1]['finder_url'],
            reverse(
                'projectroles:roles',
                kwargs={'project': self.category.sodar_uuid},
            ),
        )

    def test_get_finder_parent(self):
        """Test GET with finder role inside parent"""
        user_finder = self.make_user('user_finder')
        self.make_assignment(self.category, user_finder, self.role_finder)
        with self.login(user_finder):
            response = self.client.get(
                self.url + '?parent=' + str(self.category.sodar_uuid),
            )
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertEqual(len(pd), 1)
        self.assertEqual(pd[0]['full_title'], self.project.title)
        self.assertEqual(pd[0]['access'], False)

    def test_get_finder_nested(self):
        """Test GET with finder role and nested access"""
        sub_cat = self.make_project(
            'SubCategory', PROJECT_TYPE_CATEGORY, self.category
        )
        self.make_assignment(sub_cat, self.user_owner, self.role_owner)
        sub_project = self.make_project(
            'SubProject', PROJECT_TYPE_PROJECT, sub_cat
        )
        self.make_assignment(sub_project, self.user_owner, self.role_owner)
        user_finder = self.make_user('user_finder')
        self.make_assignment(sub_cat, user_finder, self.role_finder)
        with self.login(user_finder):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertEqual(len(pd), 3)
        self.assertEqual(pd[0]['full_title'], self.category.full_title)
        self.assertEqual(pd[0]['access'], True)
        self.assertEqual(pd[1]['full_title'], sub_cat.full_title)
        self.assertEqual(pd[1]['access'], True)
        self.assertEqual(pd[2]['full_title'], sub_project.full_title)
        self.assertEqual(pd[2]['access'], False)

    def test_get_finder_other_branch(self):
        """Test project list with finder role in another branch"""
        # Give regular access to self.project
        user_finder = self.make_user('user_finder')
        self.make_assignment(self.project, user_finder, self.role_guest)
        branch_cat = self.make_project(
            'TextCategory2', PROJECT_TYPE_CATEGORY, None
        )
        self.make_assignment(branch_cat, self.user_owner, self.role_owner)
        branch_project = self.make_project(
            'BranchProject', PROJECT_TYPE_PROJECT, branch_cat
        )
        self.make_assignment(branch_project, self.user_owner, self.role_owner)
        self.make_assignment(branch_cat, user_finder, self.role_finder)
        with self.login(user_finder):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertEqual(len(pd), 4)
        self.assertEqual(pd[0]['full_title'], self.category.full_title)
        self.assertEqual(pd[0]['access'], True)
        self.assertEqual(pd[1]['full_title'], self.project.full_title)
        self.assertEqual(pd[1]['access'], True)
        self.assertEqual(pd[2]['full_title'], branch_cat.full_title)
        self.assertEqual(pd[2]['access'], True)
        self.assertEqual(pd[3]['full_title'], branch_project.full_title)
        self.assertEqual(pd[3]['access'], False)

    def test_get_starred(self):
        """Test GET with starred project"""
        app_settings.set(
            APP_NAME, 'project_star', True, project=self.project, user=self.user
        )
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['projects'][1]['starred'], True)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive(True)
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['projects'][1]['archive'], True)

    def test_get_public_access(self):
        """Test GET with public read-only access"""
        self.project.public_access = self.role_guest
        self.project.save()
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertEqual(len(pd), 2)
        self.assertEqual(pd[1]['public_access'], True)
        self.assertEqual(pd[1]['access'], True)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_public_access_anon(self):
        """Test GET with public read-only access and anonymous user"""
        self.project.public_access = self.role_guest
        self.project.save()
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertEqual(len(pd), 2)
        self.assertEqual(pd[1]['public_access'], True)
        self.assertEqual(pd[1]['access'], True)

    def test_get_category_public_stats(self):
        """Test GET with category public stats"""
        app_settings.set(
            APP_NAME, 'category_public_stats', True, project=self.category
        )
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertEqual(len(pd), 2)
        self.assertEqual(pd[0]['full_title'], self.category.full_title)
        self.assertEqual(pd[0]['access'], True)
        self.assertEqual(pd[0]['public_access'], False)
        self.assertEqual(pd[0]['public_stats'], True)

    def test_get_category_public_stats_no_roles(self):
        """Test GET with category public stats as user with no roles"""
        app_settings.set(
            APP_NAME, 'category_public_stats', True, project=self.category
        )
        with self.login(self.user_no_roles):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertEqual(len(pd), 1)
        self.assertEqual(pd[0]['full_title'], self.category.full_title)
        self.assertEqual(pd[0]['access'], True)
        self.assertEqual(pd[0]['public_stats'], True)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_category_public_stats_anon(self):
        """Test GET with category public stats as anonymous user"""
        app_settings.set(
            APP_NAME, 'category_public_stats', True, project=self.category
        )
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        pd = response.data['projects']
        self.assertEqual(len(pd), 1)
        self.assertEqual(pd[0]['full_title'], self.category.full_title)
        self.assertEqual(pd[0]['access'], True)
        self.assertEqual(pd[0]['public_stats'], True)


class TestProjectListColumnAjaxView(
    ProjectMixin, RoleAssignmentMixin, ViewTestBase
):
    """Tests for ProjectListColumnAjaxView"""

    def setUp(self):
        super().setUp()
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user, self.role_owner
        )

    def test_post(self):
        """Test POST for custom column retrieval"""
        with self.login(self.user):
            response = self.client.post(
                reverse('projectroles:ajax_project_list_columns'),
                json.dumps({'projects': [str(self.project.sodar_uuid)]}),
                content_type='application/json',
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            str(self.project.sodar_uuid): {
                APP_NAME_FF: {'files': {'html': '0'}, 'links': {'html': '0'}}
            }
        }
        self.assertEqual(response.data, expected)

    @override_settings(FILESFOLDERS_SHOW_LIST_COLUMNS=False)
    def test_post_no_columns(self):
        """Test POST with no custom colums"""
        with self.login(self.user):
            response = self.client.post(
                reverse('projectroles:ajax_project_list_columns'),
                json.dumps({'projects': [str(self.project.sodar_uuid)]}),
                content_type='application/json',
            )
        self.assertEqual(response.status_code, 200)
        expected = {str(self.project.sodar_uuid): {}}
        self.assertEqual(response.data, expected)

    def test_post_no_permission(self):
        """Test POST with no user permission on a project"""
        new_project = self.make_project(
            'NewProject', PROJECT_TYPE_PROJECT, None
        )
        new_user = self.make_user('new_user')
        self.make_assignment(new_project, new_user, self.role_owner)

        with self.login(new_user):
            response = self.client.post(
                reverse('projectroles:ajax_project_list_columns'),
                json.dumps(
                    {
                        'projects': [
                            str(self.project.sodar_uuid),
                            str(new_project.sodar_uuid),
                        ]
                    }
                ),
                content_type='application/json',
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            str(new_project.sodar_uuid): {
                APP_NAME_FF: {'files': {'html': '0'}, 'links': {'html': '0'}}
            }
        }
        self.assertEqual(response.data, expected)


class TestProjectListRoleAjaxView(
    ProjectMixin, RoleAssignmentMixin, ViewTestBase
):
    """Tests for ProjectListRoleAjaxView"""

    def setUp(self):
        super().setUp()
        self.user_owner_cat = self.make_user('user_owner_cat')
        self.user_owner = self.make_user('user_owner')
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user_owner_cat, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )

    def test_post_category_owner(self):
        """Test POST for role retrieval as category owner"""
        with self.login(self.user_owner_cat):
            response = self.client.post(
                reverse('projectroles:ajax_project_list_roles'),
                json.dumps(
                    {
                        'projects': [
                            str(self.category.sodar_uuid),
                            str(self.project.sodar_uuid),
                        ]
                    }
                ),
                content_type='application/json',
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            str(self.category.sodar_uuid): {'name': 'Owner', 'class': None},
            str(self.project.sodar_uuid): {'name': 'Owner', 'class': None},
        }
        self.assertEqual(response.data, expected)

    def test_post_project_owner(self):
        """Test POST for role retrieval as project owner"""
        with self.login(self.user_owner):
            response = self.client.post(
                reverse('projectroles:ajax_project_list_roles'),
                json.dumps(
                    {
                        'projects': [
                            str(self.category.sodar_uuid),
                            str(self.project.sodar_uuid),
                        ]
                    }
                ),
                content_type='application/json',
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            str(self.category.sodar_uuid): {
                'name': 'N/A',
                'class': 'text-muted',
            },
            str(self.project.sodar_uuid): {'name': 'Owner', 'class': None},
        }
        self.assertEqual(response.data, expected)

    def test_post_no_access(self):
        """Test POST for role retrieval with no access to project"""
        new_user = self.make_user('new_user')  # User with no roles
        with self.login(new_user):
            response = self.client.post(
                reverse('projectroles:ajax_project_list_roles'),
                json.dumps(
                    {
                        'projects': [
                            str(self.category.sodar_uuid),
                            str(self.project.sodar_uuid),
                        ]
                    }
                ),
                content_type='application/json',
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {})


class TestCategoryStatisticsAjaxView(
    ProjectMixin, RoleAssignmentMixin, ViewTestBase
):
    """Tests for CategoryStatisticsAjaxView"""

    def setUp(self):
        super().setUp()
        self.user_owner = self.make_user('user_owner')
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user_owner, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )

    def test_get(self):
        """Test CategoryStatisticsAjaxView GET"""
        with self.login(self.user_owner):
            response = self.client.get(
                reverse(
                    'projectroles:ajax_stats_category',
                    kwargs={'project': self.category.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 200)
        expected = [
            {
                'title': 'Projects',
                'value': 1,
                'unit': None,
                'description': 'Projects in this category',
                'icon': 'mdi:cube',
            },
            {
                'title': 'Members',
                'value': 1,
                'unit': None,
                'description': 'Users with roles in this category',
                'icon': 'mdi:account-multiple',
            },
            {
                'title': 'Files',
                'value': 0,
                'unit': None,
                'description': 'Files uploaded to projects in this category',
                'icon': 'mdi:file',
            },
        ]
        self.assertEqual(response.data['stats'], expected)

    def test_get_project(self):
        """Test GET with project (should fail)"""
        with self.login(self.user_owner):
            response = self.client.get(
                reverse(
                    'projectroles:ajax_stats_category',
                    kwargs={'project': self.project.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 403)

    def test_get_subcategory(self):
        """Test GET with subcategory"""
        new_category = self.make_project(
            'NewCategory', PROJECT_TYPE_CATEGORY, self.category
        )
        self.make_assignment(new_category, self.user_owner, self.role_owner)
        with self.login(self.user_owner):
            response = self.client.get(
                reverse(
                    'projectroles:ajax_stats_category',
                    kwargs={'project': new_category.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['stats'][0]['value'], 0)
        self.assertEqual(response.data['stats'][1]['value'], 1)  # User exists
        self.assertEqual(response.data['stats'][2]['value'], 0)


class TestProjectStarringAjaxView(
    ProjectMixin, RoleAssignmentMixin, ViewTestBase
):
    """Tests for ProjectStarringAjaxView"""

    def _assert_setting_count(self, count: int):
        qs = AppSetting.objects.filter(
            app_plugin=None, project=self.project, name='project_star'
        )
        self.assertEqual(qs.count(), count)

    def setUp(self):
        super().setUp()
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, None
        )
        self.owner_as = self.make_assignment(
            self.project, self.user, self.role_owner
        )
        self.url = reverse(
            'projectroles:ajax_star',
            kwargs={'project': self.project.sodar_uuid},
        )

    def test_post(self):
        """Test ProjectStarringAjaxView POST"""
        self._assert_setting_count(0)
        with self.login(self.user):
            response = self.client.post(self.url)
        self.assertEqual(response.status_code, 200)
        self._assert_setting_count(1)
        self.assertEqual(
            app_settings.get(APP_NAME, 'project_star', self.project, self.user),
            True,
        )

    def test_post_update(self):
        """Test POST to update existing value"""
        app_settings.set(
            APP_NAME, 'project_star', True, self.project, self.user
        )
        self._assert_setting_count(1)
        self.assertEqual(
            app_settings.get(APP_NAME, 'project_star', self.project, self.user),
            True,
        )
        with self.login(self.user):
            response = self.client.post(self.url)
        self.assertEqual(response.status_code, 200)
        self._assert_setting_count(1)
        self.assertEqual(
            app_settings.get(APP_NAME, 'project_star', self.project, self.user),
            False,
        )


class TestHomeStarringAjaxView(ViewTestBase):
    """Tests for HomeStarringAjaxView"""

    def setUp(self):
        super().setUp()
        self.url = reverse('projectroles:ajax_star_home')

    def test_post(self):
        """Test HomeStarringAjaxView POST"""
        self.assertEqual(
            app_settings.get(
                APP_NAME, 'project_list_home_starred', user=self.user
            ),
            False,
        )
        with self.login(self.user):
            response = self.client.post(self.url, data={'value': '1'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            app_settings.get(
                APP_NAME, 'project_list_home_starred', user=self.user
            ),
            True,
        )

    def test_post_update(self):
        """Test POST to update existing value"""
        app_settings.set(
            APP_NAME, 'project_list_home_starred', True, user=self.user
        )
        self.assertEqual(
            app_settings.get(
                APP_NAME, 'project_list_home_starred', user=self.user
            ),
            True,
        )
        with self.login(self.user):
            response = self.client.post(self.url, data={'value': '0'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            app_settings.get(
                APP_NAME, 'project_list_home_starred', user=self.user
            ),
            False,
        )

    def test_post_value_unchanged(self):
        """Test POST with unchanged value"""
        self.assertEqual(
            app_settings.get(
                APP_NAME, 'project_list_home_starred', user=self.user
            ),
            False,
        )
        with self.login(self.user):
            response = self.client.post(self.url, data={'value': '0'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            app_settings.get(
                APP_NAME, 'project_list_home_starred', user=self.user
            ),
            False,
        )


class TestRemoteProjectAccessAjaxView(
    ProjectMixin,
    RoleAssignmentMixin,
    RemoteSiteMixin,
    RemoteProjectMixin,
    ViewTestBase,
):
    """Tests for RemoteProjectAccessAjaxView"""

    @classmethod
    def _get_query_string(cls, *args) -> str:
        """Get query string for GET requests"""
        return '?' + '&'.join(['rp={}'.format(rp.sodar_uuid) for rp in args])

    def setUp(self):
        super().setUp()
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user, self.role_owner
        )
        self.remote_site = self.make_site(
            name='RemoteSite',
            url='https://remote.site',
            mode=SITE_MODE_TARGET,
        )
        self.remote_project = self.make_remote_project(
            project_uuid=self.project.sodar_uuid,
            site=self.remote_site,
            level=REMOTE_LEVEL_READ_ROLES,
            project=self.project,
        )
        self.rp_uuid = str(self.remote_project.sodar_uuid)
        self.remote_site2 = self.make_site(
            name='RemoteSite2',
            url='https://remote.site2',
            mode=SITE_MODE_TARGET,
        )
        self.remote_project2 = self.make_remote_project(
            project_uuid=self.project.sodar_uuid,
            site=self.remote_site2,
            level=REMOTE_LEVEL_READ_ROLES,
            project=self.project,
        )
        self.rp_uuid2 = str(self.remote_project2.sodar_uuid)
        self.query_string = self._get_query_string(
            self.remote_project, self.remote_project2
        )
        self.url = reverse(
            'projectroles:ajax_remote_access',
            kwargs={'project': self.project.sodar_uuid},
        )

    def test_get(self):
        """Test RemoteProjectAccessAjaxView GET"""
        with self.login(self.user):
            response = self.client.get(self.url + self.query_string)
        self.assertEqual(response.status_code, 200)
        expected = {self.rp_uuid: False, self.rp_uuid2: False}
        self.assertEqual(response.data, expected)

    def test_get_accessed(self):
        """Test GET with one remote project accessed"""
        self.remote_project.date_access = timezone.now()
        self.remote_project.save()
        with self.login(self.user):
            response = self.client.get(self.url + self.query_string)
        self.assertEqual(response.status_code, 200)
        expected = {self.rp_uuid: True, self.rp_uuid2: False}
        self.assertEqual(response.data, expected)

    def test_get_accessed_multiple(self):
        """Test GET with multiple remote projects accessed"""
        self.remote_project.date_access = timezone.now()
        self.remote_project.save()
        self.remote_project2.date_access = timezone.now()
        self.remote_project2.save()
        with self.login(self.user):
            response = self.client.get(self.url + self.query_string)
        self.assertEqual(response.status_code, 200)
        expected = {self.rp_uuid: True, self.rp_uuid2: True}
        self.assertEqual(response.data, expected)

    def test_get_wrong_project(self):
        """Test GET with remote project for a different project"""
        new_project = self.make_project(
            'NewProject', PROJECT_TYPE_PROJECT, self.category
        )
        new_rp = self.make_remote_project(
            new_project.sodar_uuid,
            self.remote_site,
            REMOTE_LEVEL_READ_ROLES,
            project=new_project,
        )
        with self.login(self.user):
            response = self.client.get(
                self.url + self._get_query_string(new_rp)
            )
        self.assertEqual(response.status_code, 400)

    def test_get_invalid_uuid(self):
        """Test GET with invalid remote project UUID"""
        with self.login(self.user):
            response = self.client.get(self.url + '?rp=' + INVALID_UUID)
        self.assertEqual(response.status_code, 404)


class TestSidebarContentAjaxView(
    ProjectMixin, RoleAssignmentMixin, ViewTestBase
):
    """Tests for SidebarContentAjaxView"""

    def setUp(self):
        super().setUp()
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user, self.role_owner
        )

    def test_get(self):
        """Test SidebarContentAjaxView GET"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'projectroles:ajax_sidebar',
                    kwargs={'project': self.project.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            'links': app_links.get_project_links(self.user, self.project)
        }
        self.assertEqual(response.json(), expected)

    def test_get_category(self):
        """Test GET with category"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'projectroles:ajax_sidebar',
                    kwargs={'project': self.category.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            'links': app_links.get_project_links(self.user, self.category)
        }
        self.assertEqual(response.json(), expected)

    def test_get_app_link(self):
        """Test GET with app plugin link"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'projectroles:ajax_sidebar',
                    kwargs={'project': self.project.sodar_uuid},
                )
                + '?app_name=filesfolders'
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            'links': app_links.get_project_links(
                self.user, self.project, app_name=APP_NAME_FF
            )
        }
        self.assertEqual(response.json(), expected)

    def test_get_url_name(self):
        """Test GET with URL name"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'projectroles:ajax_sidebar',
                    kwargs={'project': self.project.sodar_uuid},
                )
                + '?app_name=filesfolders&url_name=file_create'
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            'links': app_links.get_project_links(
                self.user,
                self.project,
                app_name=APP_NAME_FF,
                url_name='file_create',
            )
        }
        self.assertEqual(response.json(), expected)


class TestSiteReadOnlySettingAjaxView(SerializedObjectMixin, TestCase):
    """Tests for SiteReadOnlySettingAjaxView"""

    def setUp(self):
        self.user = self.make_user('user')
        self.user.save()

    def test_get_disabled(self):
        """Test SiteReadOnlySettingAjaxView GET with read-only mode disabled"""
        self.assertFalse(app_settings.get(APP_NAME, 'site_read_only'))
        with self.login(self.user):
            response = self.client.get(
                reverse('projectroles:ajax_settings_site_read_only')
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {'site_read_only': False})

    def test_get_enabled(self):
        """Test GET with read-only mode enabled"""
        app_settings.set(APP_NAME, 'site_read_only', True)
        with self.login(self.user):
            response = self.client.get(
                reverse('projectroles:ajax_settings_site_read_only')
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {'site_read_only': True})


class TestUserDropdownContentAjaxView(ViewTestBase):
    """Tests for UserDropdownContentAjaxView"""

    def setUp(self):
        super().setUp()
        self.user = self.make_user('user')
        self.user.is_superuser = True
        self.user.save()
        self.reg_user = self.make_user('reg_user')

    def test_get(self):
        """Test UserDropdownContentAjaxView GET as regular user"""
        with self.login(self.reg_user):
            response = self.client.get(
                reverse('projectroles:ajax_user_dropdown')
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(), {'links': app_links.get_user_links(self.reg_user)}
        )

    def test_get_superuser(self):
        """Test GET as superuser"""
        with self.login(self.user):
            response = self.client.get(
                reverse('projectroles:ajax_user_dropdown')
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(), {'links': app_links.get_user_links(self.user)}
        )

    def test_get_app_name(self):
        """Test GET with app plugin name"""
        with self.login(self.reg_user):
            response = self.client.get(
                reverse('projectroles:ajax_user_dropdown') + '?app_name=tokens'
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                'links': app_links.get_user_links(
                    self.reg_user, app_name='tokens'
                )
            },
        )

    def test_get_url_name(self):
        """Test GET with URL name"""
        with self.login(self.reg_user):
            response = self.client.get(
                reverse('projectroles:ajax_user_dropdown')
                + '?app_name=tokens&url_name=create'
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                'links': app_links.get_user_links(
                    self.reg_user, app_name='tokens', url_name='create'
                )
            },
        )


class TestCurrentUserRetrieveAjaxView(SerializedObjectMixin, TestCase):
    """Tests for CurrentUserRetrieveAjaxView"""

    def setUp(self):
        self.user = self.make_user('user')
        self.user.is_superuser = True
        self.user.save()
        self.reg_user = self.make_user('reg_user')

    def test_get_regular_user(self):
        """Test CurrentUserRetrieveAjaxView GET with regular user"""
        with self.login(self.reg_user):
            response = self.client.get(
                reverse('projectroles:ajax_user_current')
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, self.get_serialized_user(self.reg_user))

    def test_get_superuser(self):
        """Test GET with superuser"""
        with self.login(self.user):
            response = self.client.get(
                reverse('projectroles:ajax_user_current')
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, self.get_serialized_user(self.user))


class TestUserAutocompleteAjaxView(
    ProjectMixin, RoleAssignmentMixin, ViewTestBase
):
    """Tests for UserAutocompleteAjaxView"""

    def _get_forward_qs(self, data: dict) -> str:
        """Return forward querystring for UserAutocompleteAjaxView requests"""
        return '?forward=' + json.dumps(data)

    def setUp(self):
        super().setUp()
        self.user_owner = self.make_user('user_owner')
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user_owner, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        self.user_no_roles = self.make_user('user_no_roles')
        self.url = reverse('projectroles:ajax_autocomplete_user')

    def test_get(self):
        """Test UserAutocompleteAjaxView GET with default params"""
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['results']), 3)

    def test_get_project(self):
        """Test GET with project scope"""
        data = {
            'project': str(self.project.sodar_uuid),
            'scope': 'project',
        }
        with self.login(self.user):
            response = self.client.get(self.url + self._get_forward_qs(data))
        self.assertEqual(response.status_code, 200)
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['id'], str(self.user_owner.sodar_uuid))

    def test_get_project_exclude(self):
        """Test GET with project_exclude scope"""
        data = {
            'project': str(self.project.sodar_uuid),
            'scope': 'project_exclude',
        }
        with self.login(self.user):
            response = self.client.get(self.url + self._get_forward_qs(data))
        self.assertEqual(response.status_code, 200)
        results = response.json()['results']
        self.assertEqual(len(results), 2)
        self.assertNotIn(
            str(self.user_owner.sodar_uuid), [r['id'] for r in results]
        )

    def test_get_inactive_user(self):
        """Test GET with inactive user"""
        self.user_owner.is_active = False
        self.user_owner.save()
        data = {
            'project': str(self.project.sodar_uuid),
            'scope': 'project',
        }
        with self.login(self.user):
            response = self.client.get(self.url + self._get_forward_qs(data))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['results']), 0)

    @override_settings(PROJECTROLES_ALLOW_LOCAL_USERS=False)
    def test_get_disallow_local(self):
        """Test GET with local users disallowed"""
        with self.login(self.user_owner):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['results']), 0)

    @override_settings(PROJECTROLES_ALLOW_LOCAL_USERS=False)
    def test_get_disallow_local_as_superuser(self):
        """Test GET with local users disallowed as superuser"""
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        # For superuser, all users should be returned
        self.assertEqual(len(response.json()['results']), 3)

    def test_get_user_exclude(self):
        """Test GET with user excluding"""
        data = {
            'scope': 'all',
            'exclude': [str(self.user_owner.sodar_uuid)],
        }
        with self.login(self.user):
            response = self.client.get(self.url + self._get_forward_qs(data))
        self.assertEqual(response.status_code, 200)
        results = response.json()['results']
        self.assertEqual(len(results), 2)
        self.assertNotIn(
            str(self.user_owner.sodar_uuid), [r['id'] for r in results]
        )


class TestUserAutocompleteRedirectAjaxView(
    ProjectMixin, RoleAssignmentMixin, ViewTestBase
):
    """Tests for UserAutocompleteRedirectAjaxView"""

    def setUp(self):
        super().setUp()
        self.user_owner = self.make_user('user_owner')
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user_owner, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        self.url = reverse('projectroles:ajax_autocomplete_user_redirect')

    def test_get(self):
        """Test TestUserAutocompleteRedirectAjaxView GET"""
        data = {
            'project': self.project.sodar_uuid,
            'role': self.role_guest.pk,
            'q': 'test@example.com',
        }
        with self.login(self.user):
            response = self.client.get(self.url, data)
        self.assertEqual(response.status_code, 200)
        option_new = {
            'id': 'test@example.com',
            'text': 'Send an invite to "test@example.com"',
            'create_id': True,
        }
        data = json.loads(response.content)
        self.assertIn(option_new, data['results'])

    def test_get_invalid_email(self):
        """Test GET with invalid email"""
        data = {
            'project': self.project.sodar_uuid,
            'role': self.role_guest.pk,
            'q': 'test@example',
        }
        with self.login(self.user):
            response = self.client.get(self.url, data)
        self.assertEqual(response.status_code, 200)
        option_new = {
            'id': 'test@example.com',
            'text': 'Send an invite to "test@example"',
            'create_id': True,
        }
        data = json.loads(response.content)
        self.assertNotIn(option_new, data['results'])

    def test_post_redirect_to_invite(self):
        """Test POST for redirecting to ProjectInviteCreateView"""
        data = {
            'project': self.project.sodar_uuid,
            'role': self.role_guest.pk,
            'text': 'test@example.com',
        }
        with self.login(self.user):
            response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(data['success'], True)
        self.assertEqual(
            data['redirect_url'],
            reverse(
                'projectroles:invite_create',
                kwargs={'project': self.project.sodar_uuid},
            ),
        )



"""Access rules for the timeline app"""

import rules

# Projectroles dependency
from projectroles import rules as pr_rules  # To access common predicates


# Predicates -------------------------------------------------------------


# If we need to assign new predicates, we do it here


# Rules ------------------------------------------------------------------


# Rules should not be needed, use permissions for user rights


# Permissions ------------------------------------------------------------


# Allow viewing project timeline
rules.add_perm(
    'timeline.view_timeline',
    pr_rules.is_project_owner
    | pr_rules.is_project_delegate
    | pr_rules.is_project_contributor
    | pr_rules.is_project_guest,
)

# Allow viewing timeline for site-specific events
rules.add_perm('timeline.view_site_timeline', rules.is_authenticated)

# Allow viewing timeline for admin site
rules.add_perm('timeline.view_site_admin', rules.is_superuser)

# Allow viewing classified event
rules.add_perm(
    'timeline.view_classified_event',
    pr_rules.is_project_owner | pr_rules.is_project_delegate,
)

# Allow viewing classified site-wide event
rules.add_perm('timeline.view_classified_site_event', rules.is_superuser)

# Allow viewing extra data of project event
rules.add_perm(
    'timeline.view_event_extra_data',
    pr_rules.is_project_owner | pr_rules.is_project_delegate,
)



"""UI views for the userprofile app"""

from django.conf import settings
from django.contrib import auth, messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import redirect
from django.urls import reverse, reverse_lazy
from django.views.generic import (
    TemplateView,
    FormView,
    CreateView,
    DeleteView,
    View,
)

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.email import send_generic_mail, get_email_user
from projectroles.models import SODARUserAdditionalEmail, SODAR_CONSTANTS
from projectroles.plugins import PluginAPI
from projectroles.views import (
    LoggedInPermissionMixin,
    HTTPRefererMixin,
    InvalidFormMixin,
    CurrentUserFormMixin,
)

from userprofile.forms import UserAppSettingsForm, UserEmailForm


app_settings = AppSettingAPI()
plugin_api = PluginAPI()
User = auth.get_user_model()


# SODAR Constants
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']

# Local Constants
APP_NAME = 'userprofile'
APP_NAME_PR = 'projectroles'
SETTING_UPDATE_MSG = 'User settings updated.'
VERIFY_EMAIL_SUBJECT = 'Verify your additional email address for {site}'
VERIFY_EMAIL_BODY = r'''
{user} has added this address to their
additional emails on {site}.

Once verified, the site may use this email to send automated email for
notifications as configured in user settings.

Please verify this email address by following this URL:
{url}

If this was not requested by you, this message can be ignored.
'''.lstrip()
EMAIL_NOT_FOUND_MSG = 'No email found.'
EMAIL_ALREADY_VERIFIED_MSG = 'Email already verified.'
EMAIL_VERIFIED_MSG = 'Email "{email}" verified.'
EMAIL_VERIFY_RESEND_MSG = 'Verification message to "{email}" resent.'


class UserDetailView(LoginRequiredMixin, LoggedInPermissionMixin, TemplateView):
    """Display the user profile view including the user settings"""

    template_name = 'userprofile/detail.html'
    permission_required = 'userprofile.view_detail'

    def _get_user_settings(self):
        """Return user setting values"""
        plugins = plugin_api.get_active_plugins(
            plugin_type='project_app'
        ) + plugin_api.get_active_plugins(plugin_type='site_app')
        for plugin in plugins + [None]:
            if plugin:
                name = plugin.name
                s_defs = app_settings.get_definitions(
                    APP_SETTING_SCOPE_USER, plugin=plugin, user_modifiable=True
                )
            else:
                name = 'projectroles'
                s_defs = app_settings.get_definitions(
                    APP_SETTING_SCOPE_USER,
                    plugin_name=name,
                    user_modifiable=True,
                )
            for s_def in s_defs.values():
                yield {
                    'label': s_def.label or f'{name}.{s_def.name}',
                    'value': app_settings.get(
                        name, s_def.name, user=self.request.user
                    ),
                    'description': s_def.description,
                }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['user_settings'] = list(self._get_user_settings())
        context['add_emails'] = SODARUserAdditionalEmail.objects.filter(
            user=self.request.user
        ).order_by('email')
        context['site_read_only'] = app_settings.get(
            APP_NAME_PR, 'site_read_only'
        )
        context['send_email'] = settings.PROJECTROLES_SEND_EMAIL
        context['site_mode'] = settings.PROJECTROLES_SITE_MODE
        return context


class UserAppSettingsView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    HTTPRefererMixin,
    InvalidFormMixin,
    FormView,
):
    """User app settings form view"""

    form_class = UserAppSettingsForm
    permission_required = 'userprofile.update_settings'
    template_name = 'userprofile/settings_form.html'
    success_url = reverse_lazy('userprofile:detail')

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['current_user'] = self.request.user
        return kwargs

    def form_valid(self, form):
        result = super().form_valid(form)
        for k, v in form.cleaned_data.items():
            if k.startswith('settings.'):
                _, plugin_name, setting_name = k.split('.', 3)
                # TODO: Omit global USER settings (#1329)
                app_settings.set(
                    plugin_name, setting_name, v, user=self.request.user
                )
        messages.success(self.request, SETTING_UPDATE_MSG)
        return result


class UserEmailMixin:
    """Mixin for user email helpers"""

    def send_verify_email(
        self, email: SODARUserAdditionalEmail, resend: bool = False
    ):
        """
        Send verification message to additional email address.

        :param email: SODARUserAdditionalEmail object
        """
        subject = VERIFY_EMAIL_SUBJECT.format(site=settings.SITE_INSTANCE_TITLE)
        body = VERIFY_EMAIL_BODY.format(
            user=get_email_user(email.user),
            site=settings.SITE_INSTANCE_TITLE,
            url=self.request.build_absolute_uri(
                reverse(
                    'userprofile:email_verify', kwargs={'secret': email.secret}
                )
            ),
        )
        try:
            send_generic_mail(subject, body, [email.email], self.request)
            if resend:
                messages.success(
                    self.request,
                    EMAIL_VERIFY_RESEND_MSG.format(email=email.email),
                )
            else:
                messages.success(
                    self.request,
                    'Email added. A verification message has been sent to the '
                    'address. Follow the received verification link to '
                    'activate the address.',
                )
        except Exception as ex:
            messages.error(
                self.request, f'Failed to send verification mail: {ex}'
            )


class UserEmailCreateView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    CurrentUserFormMixin,
    UserEmailMixin,
    CreateView,
):
    """User additional email creation view"""

    form_class = UserEmailForm
    permission_required = 'userprofile.create_email'
    template_name = 'userprofile/email_form.html'

    def get_success_url(self):
        timeline = plugin_api.get_backend_api('timeline_backend')
        self.send_verify_email(self.object)
        if timeline:
            timeline.add_event(
                project=None,
                app_name=APP_NAME,
                user=self.request.user,
                event_name='email_create',
                description=f'create additional email "{self.object.email}"',
                classified=True,
                status_type=timeline.TL_STATUS_OK,
            )
        return reverse('userprofile:detail')

    def get(self, *args, **kwargs):
        if not settings.PROJECTROLES_SEND_EMAIL:
            messages.warning(
                self.request,
                'Email sending disabled, adding email addresses is not '
                'allowed.',
            )
            return redirect(reverse('userprofile:detail'))
        return super().get(*args, **kwargs)


class UserEmailVerifyView(LoginRequiredMixin, LoggedInPermissionMixin, View):
    """View for verifying a created additional email address"""

    http_method_names = ['get']
    permission_required = 'userprofile.create_email'

    def get(self, request, *args, **kwargs):
        secret = self.kwargs.get('secret')
        email = SODARUserAdditionalEmail.objects.filter(
            user=request.user, secret=secret
        ).first()
        if not email:
            messages.error(request, EMAIL_NOT_FOUND_MSG)
        elif email.verified:
            messages.info(request, EMAIL_ALREADY_VERIFIED_MSG)
        else:
            email.verified = True
            email.save()
            messages.success(
                request, EMAIL_VERIFIED_MSG.format(email=email.email)
            )
        return redirect(reverse('userprofile:detail'))


class UserEmailVerifyResendView(
    LoginRequiredMixin, LoggedInPermissionMixin, UserEmailMixin, View
):
    """View for resending additional email verification message"""

    http_method_names = ['get']
    permission_required = 'userprofile.create_email'

    def get(self, request, *args, **kwargs):
        email_uuid = self.kwargs.get('sodaruseradditionalemail')
        email = SODARUserAdditionalEmail.objects.filter(
            user=request.user, sodar_uuid=email_uuid
        ).first()
        if not email:
            messages.error(request, EMAIL_NOT_FOUND_MSG)
        elif email.verified:
            messages.info(request, EMAIL_ALREADY_VERIFIED_MSG)
        else:
            self.send_verify_email(email, resend=True)
        return redirect(reverse('userprofile:detail'))


class UserEmailDeleteView(
    LoginRequiredMixin, LoggedInPermissionMixin, DeleteView
):
    """View for deleting additional email"""

    model = SODARUserAdditionalEmail
    permission_required = 'userprofile.delete_email'
    slug_url_kwarg = 'sodaruseradditionalemail'
    slug_field = 'sodar_uuid'
    template_name = 'userprofile/email_confirm_delete.html'

    def get_success_url(self):
        messages.success(self.request, 'Email address deleted.')
        return reverse('userprofile:detail')



"""REST API view tests for the timeline app"""

import json
import uuid

from django.urls import reverse

from test_plus.test import APITestCase

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
)
from projectroles.tests.test_views_api import (
    SODARAPIViewTestMixin,
    TEST_SERVER_URL,
)

from timeline.tests.test_models import (
    TimelineEventMixin,
    TimelineEventStatusMixin,
    TimelineEventObjectRefMixin,
    EXTRA_DATA,
)
from timeline.views_api import (
    TIMELINE_API_MEDIA_TYPE,
    TIMELINE_API_DEFAULT_VERSION,
)


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']

# Local constants
APP_NAME_PR = 'projectroles'
EVENT_NAME = 'test_event'
EVENT_NAME_CLASSIFIED = 'test_event_classified'
EVENT_DESC = 'description'
OBJ_REF_LABEL = 'test_label'
OBJ_REF_NAME = 'test_obj_name'


class TimelineAPIViewTestBase(
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    TimelineEventMixin,
    TimelineEventStatusMixin,
    TimelineEventObjectRefMixin,
    SODARAPIViewTestMixin,
    APITestCase,
):
    """Base class for timeline API view tests"""

    media_type = TIMELINE_API_MEDIA_TYPE
    api_version = TIMELINE_API_DEFAULT_VERSION

    def setUp(self):
        self.maxDiff = None
        # Init roles
        self.init_roles()
        # Init users
        self.superuser = self.make_user('superuser')
        self.superuser.is_staff = True
        self.superuser.is_superuser = True
        self.superuser.save()
        self.user_owner = self.make_user('user_owner')
        self.user_delegate = self.make_user('user_delegate')
        self.user_contributor = self.make_user('user_contributor')
        # Init category and project
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user_owner, self.role_owner
        )
        # Init project
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        self.delegate_as = self.make_assignment(
            self.project, self.user_delegate, self.role_delegate
        )
        self.contrib_as = self.make_assignment(
            self.project, self.user_contributor, self.role_contributor
        )
        # Get knox token for self.user
        self.knox_token = self.get_token(self.superuser)


class TestProjectTimelineEventListAPIView(TimelineAPIViewTestBase):
    """Tests for ProjectTimelineEventListAPIView"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            project=self.project,
            app=APP_NAME_PR,
            user=self.user_owner,
            event_name=EVENT_NAME,
            description=EVENT_DESC,
            classified=False,
            extra_data=EXTRA_DATA,
        )
        self.event_classified = self.make_event(
            project=self.project,
            app=APP_NAME_PR,
            user=self.user_owner,
            event_name=EVENT_NAME_CLASSIFIED,
            description=EVENT_DESC,
            classified=True,
            extra_data=EXTRA_DATA,
        )
        self.url = reverse(
            'timeline:api_list_project',
            kwargs={'project': self.project.sodar_uuid},
        )

    def test_get_superuser(self):
        """Test ProjectTimelineEventListAPIView GET as superuser"""
        response = self.request_knox(
            self.url, token=self.get_token(self.superuser)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertEqual(len(response_data), 2)
        expected = {
            'project': str(self.project.sodar_uuid),
            'app': APP_NAME_PR,
            'plugin': None,
            'user': str(self.user_owner.sodar_uuid),
            'event_name': EVENT_NAME,
            'description': EVENT_DESC,
            'extra_data': EXTRA_DATA,
            'classified': False,
            'status_changes': [],
            'event_objects': [],
            'sodar_uuid': str(self.event.sodar_uuid),
        }
        self.assertEqual(response_data[1], expected)
        expected = {
            'project': str(self.project.sodar_uuid),
            'app': APP_NAME_PR,
            'plugin': None,
            'user': str(self.user_owner.sodar_uuid),
            'event_name': EVENT_NAME_CLASSIFIED,
            'description': EVENT_DESC,
            'extra_data': EXTRA_DATA,
            'classified': True,
            'status_changes': [],
            'event_objects': [],
            'sodar_uuid': str(self.event_classified.sodar_uuid),
        }
        self.assertEqual(response_data[0], expected)

    def test_get_pagination(self):
        """Test GET with pagination"""
        url = self.url + '?page=1'
        response = self.request_knox(url, token=self.get_token(self.superuser))
        self.assertEqual(response.status_code, 200)
        expected = {
            'count': 2,
            'next': TEST_SERVER_URL + self.url + '?page=2',
            'previous': None,
            'results': [
                {
                    'project': str(self.project.sodar_uuid),
                    'app': APP_NAME_PR,
                    'plugin': None,
                    'user': str(self.user_owner.sodar_uuid),
                    'event_name': EVENT_NAME_CLASSIFIED,
                    'description': EVENT_DESC,
                    'extra_data': EXTRA_DATA,
                    'classified': True,
                    'status_changes': [],
                    'event_objects': [],
                    'sodar_uuid': str(self.event_classified.sodar_uuid),
                }
            ],
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_get_owner(self):
        """Test GET as owner"""
        response = self.request_knox(
            self.url, token=self.get_token(self.user_owner)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertEqual(len(response_data), 2)
        self.assertEqual(response_data[0]['extra_data'], EXTRA_DATA)
        self.assertEqual(response_data[1]['extra_data'], EXTRA_DATA)

    def test_get_delegate(self):
        """Test GET as delegate"""
        response = self.request_knox(
            self.url, token=self.get_token(self.user_delegate)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertEqual(len(response_data), 2)
        self.assertEqual(response_data[0]['extra_data'], EXTRA_DATA)
        self.assertEqual(response_data[1]['extra_data'], EXTRA_DATA)

    def test_get_contributor(self):
        """Test GET as contributor"""
        response = self.request_knox(
            self.url, token=self.get_token(self.user_contributor)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertEqual(len(response_data), 1)
        self.assertEqual(
            response_data[0]['sodar_uuid'], str(self.event.sodar_uuid)
        )
        self.assertNotIn('extra_data', response_data[0])

    def test_get_object_ref_owner(self):
        """Test GET with object reference as owner"""
        obj_ref = self.make_object_ref(
            event=self.event,
            obj=self.user_owner,
            label=OBJ_REF_LABEL,
            name=OBJ_REF_NAME,
            uuid=self.user_owner.sodar_uuid,
            extra_data=EXTRA_DATA,
        )
        response = self.request_knox(
            self.url, token=self.get_token(self.user_owner)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        expected = [
            {
                'event': str(self.event.sodar_uuid),
                'label': OBJ_REF_LABEL,
                'name': OBJ_REF_NAME,
                'object_model': 'User',
                'object_uuid': str(self.user_owner.sodar_uuid),
                'extra_data': EXTRA_DATA,
                'sodar_uuid': str(obj_ref.sodar_uuid),
            }
        ]
        self.assertEqual(response_data[1]['event_objects'], expected)

    def test_get_object_ref_delegate(self):
        """Test GET with object reference as delegate"""
        self.make_object_ref(
            event=self.event,
            obj=self.user_owner,
            label=OBJ_REF_LABEL,
            name=OBJ_REF_NAME,
            uuid=self.user_owner.sodar_uuid,
            extra_data=EXTRA_DATA,
        )
        response = self.request_knox(
            self.url, token=self.get_token(self.user_delegate)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertEqual(len(response_data[1]['event_objects']), 1)
        self.assertEqual(
            response_data[1]['event_objects'][0]['extra_data'], EXTRA_DATA
        )

    def test_get_object_ref_contributor(self):
        """Test GET with object reference as contributor"""
        self.make_object_ref(
            event=self.event,
            obj=self.user_owner,
            label=OBJ_REF_LABEL,
            name=OBJ_REF_NAME,
            uuid=self.user_owner.sodar_uuid,
            extra_data=EXTRA_DATA,
        )
        response = self.request_knox(
            self.url, token=self.get_token(self.user_contributor)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertEqual(len(response_data[0]['event_objects']), 1)
        # No extra data should be returned for contributors and below
        self.assertNotIn('extra_data', response_data[0]['event_objects'][0])

    def test_get_status_owner(self):
        """Test GET with status changes as owner"""
        status_init = self.make_event_status(
            event=self.event,
            status_type='INIT',
            description='INIT',
            extra_data=EXTRA_DATA,
        )
        status_ok = self.make_event_status(
            event=self.event,
            status_type='OK',
            description='OK',
            extra_data=EXTRA_DATA,
        )
        response = self.request_knox(
            self.url, token=self.get_token(self.user_owner)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        expected = [
            {
                'event': str(self.event.sodar_uuid),
                'timestamp': self.get_drf_datetime(status_init.timestamp),
                'status_type': 'INIT',
                'description': 'INIT',
                'extra_data': EXTRA_DATA,
                'sodar_uuid': str(status_init.sodar_uuid),
            },
            {
                'event': str(self.event.sodar_uuid),
                'timestamp': self.get_drf_datetime(status_ok.timestamp),
                'status_type': 'OK',
                'description': 'OK',
                'extra_data': EXTRA_DATA,
                'sodar_uuid': str(status_ok.sodar_uuid),
            },
        ]
        self.assertEqual(response_data[1]['status_changes'], expected)

    def test_get_status_delegate(self):
        """Test GET with status changes as delegate"""
        self.make_event_status(
            event=self.event,
            status_type='INIT',
            description='INIT',
            extra_data=EXTRA_DATA,
        )
        response = self.request_knox(
            self.url, token=self.get_token(self.user_delegate)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertEqual(
            response_data[1]['status_changes'][0]['extra_data'], EXTRA_DATA
        )

    def test_get_status_contributor(self):
        """Test GET with status changes as contributor"""
        self.make_event_status(
            event=self.event,
            status_type='INIT',
            description='INIT',
            extra_data=EXTRA_DATA,
        )
        response = self.request_knox(
            self.url, token=self.get_token(self.user_contributor)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertNotIn('extra_data', response_data[0]['status_changes'][0])


class TestSiteTimelineEventListAPIView(TimelineAPIViewTestBase):
    """Tests for SiteTimelineEventListAPIView"""

    def setUp(self):
        super().setUp()
        self.project_event = self.make_event(
            project=self.project,
            app=APP_NAME_PR,
            user=self.user_owner,
            event_name=EVENT_NAME,
            description=EVENT_DESC,
            classified=False,
            extra_data=EXTRA_DATA,
        )
        self.site_event = self.make_event(
            project=None,
            app=APP_NAME_PR,
            user=self.user_owner,
            event_name=EVENT_NAME,
            description=EVENT_DESC,
            classified=False,
            extra_data=EXTRA_DATA,
        )
        self.site_event_classified = self.make_event(
            project=None,
            app=APP_NAME_PR,
            user=self.user_owner,
            event_name=EVENT_NAME_CLASSIFIED,
            description=EVENT_DESC,
            classified=True,
            extra_data=EXTRA_DATA,
        )
        self.url = reverse('timeline:api_list_site')

    def test_get_superuser(self):
        """Test SiteTimelineEventListAPIView GET as superuser"""
        response = self.request_knox(
            self.url, token=self.get_token(self.superuser)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertEqual(len(response_data), 2)
        self.assertEqual(
            response_data[0]['sodar_uuid'],
            str(self.site_event_classified.sodar_uuid),
        )
        self.assertEqual(
            response_data[1]['sodar_uuid'], str(self.site_event.sodar_uuid)
        )

    def test_get_regular_user(self):
        """Test GET as regular_user"""
        response = self.request_knox(
            self.url, token=self.get_token(self.user_owner)
        )
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content)
        self.assertEqual(len(response_data), 1)
        self.assertEqual(
            response_data[0]['sodar_uuid'], str(self.site_event.sodar_uuid)
        )


class TestTimelineEventRetrieveAPIView(TimelineAPIViewTestBase):
    """Tests for TimelineEventRetrieveAPIView"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            project=self.project,
            app=APP_NAME_PR,
            user=self.user_owner,
            event_name=EVENT_NAME,
            description=EVENT_DESC,
            classified=False,
            extra_data=EXTRA_DATA,
        )

    def test_get(self):
        """Test TimelineEventRetrieveAPIView GET"""
        url = reverse(
            'timeline:api_retrieve',
            kwargs={'timelineevent': self.event.sodar_uuid},
        )
        response = self.request_knox(url, token=self.get_token(self.superuser))
        self.assertEqual(response.status_code, 200)
        expected = {
            'project': str(self.project.sodar_uuid),
            'app': APP_NAME_PR,
            'plugin': None,
            'user': str(self.user_owner.sodar_uuid),
            'event_name': EVENT_NAME,
            'description': EVENT_DESC,
            'extra_data': EXTRA_DATA,
            'classified': False,
            'status_changes': [],
            'event_objects': [],
            'sodar_uuid': str(self.event.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_get_not_found(self):
        """Test GET with invalid UUID"""
        invalid_uuid = uuid.uuid4()
        self.assertNotEqual(self.event.sodar_uuid, invalid_uuid)
        url = reverse(
            'timeline:api_retrieve', kwargs={'timelineevent': invalid_uuid}
        )
        response = self.request_knox(url, token=self.get_token(self.superuser))
        self.assertEqual(response.status_code, 404)



"""UI tests for the adminalerts app"""

from django.urls import reverse
from django.utils import timezone

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By

# Projectroles dependency
from projectroles.tests.test_ui import UITestBase

from adminalerts.tests.test_models import AdminAlertMixin


class AdminAlertUITestBase(AdminAlertMixin, UITestBase):
    def setUp(self):
        super().setUp()
        # Create users
        self.superuser = self.make_user('superuser', True)
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        self.regular_user = self.make_user('regular_user', False)
        # Create alert
        self.alert = self.make_alert(
            message='alert',
            user=self.superuser,
            description='description',
            active=True,
        )


class TestAlertMessage(AdminAlertUITestBase):
    """Tests for the admin alert message"""

    def setUp(self):
        super().setUp()
        self.url = reverse('home')

    def test_message(self):
        """Test visibility of alert message in home view"""
        expected = [(self.superuser, 1), (self.regular_user, 1)]
        self.assert_element_count(
            expected, self.url, 'sodar-alert-site-app', 'class'
        )
        self.assert_element_count(
            expected, self.url, 'sodar-alert-full-text-link', 'class'
        )
        self.assert_element_count(
            expected, self.url, 'sodar-alert-detail-link', 'class'
        )

    def test_message_no_description(self):
        """Test alert message without description"""
        self.alert.description = ''
        self.alert.save()
        expected = [(self.superuser, 1), (self.regular_user, 1)]
        self.assert_element_count(
            expected, self.url, 'sodar-alert-site-app', 'class'
        )
        expected = [(self.superuser, 0), (self.regular_user, 0)]
        self.assert_element_count(
            expected, self.url, 'sodar-alert-full-text-link', 'class'
        )
        self.assert_element_count(
            expected, self.url, 'sodar-alert-detail-link', 'class'
        )

    def test_message_inactive(self):
        """Test visibility of inactive alert message"""
        self.alert.active = 0
        self.alert.save()
        expected = [(self.superuser, 0), (self.regular_user, 0)]
        self.assert_element_count(
            expected, self.url, 'sodar-alert-site-app', 'class'
        )

    def test_message_expired(self):
        """Test visibility of expired alert message"""
        self.alert.date_expire = timezone.now() - timezone.timedelta(days=1)
        self.alert.save()
        expected = [(self.superuser, 0), (self.regular_user, 0)]
        self.assert_element_count(
            expected, self.url, 'sodar-alert-site-app', 'class'
        )

    def test_message_login(self):
        """Test visibility of alert in login view with auth requirement"""
        self.selenium.get(self.build_selenium_url(reverse('login')))
        with self.assertRaises(NoSuchElementException):
            self.selenium.find_element(By.CLASS_NAME, 'sodar-alert-site-app')

    def test_message_login_no_auth(self):
        """Test visibility of alert in login view without auth requirement"""
        self.alert.require_auth = False
        self.alert.save()
        self.selenium.get(self.build_selenium_url(reverse('login')))
        self.assertIsNotNone(
            self.selenium.find_element(By.CLASS_NAME, 'sodar-alert-site-app')
        )


class TestAdminAlertListView(AdminAlertUITestBase):
    """Tests for AdminAlertListView"""

    def setUp(self):
        super().setUp()
        self.url = reverse('adminalerts:list')

    def test_list_items(self):
        """Test existence of items in list"""
        expected = [(self.superuser, 1)]
        self.assert_element_count(
            expected, self.url, 'sodar-aa-alert-item', 'id'
        )

    def test_list_dropdown(self):
        """Test existence of alert dropdown in list"""
        expected = [(self.superuser, 1)]
        self.assert_element_count(
            expected, self.url, 'sodar-aa-alert-dropdown', 'class'
        )
